#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from cminx.parser import ParserErrorListener, CMakeSyntaxError


class TestErrorListener(unittest.TestCase):
    def setUp(self):
        self.listener = ParserErrorListener()

    def test_syntax_error(self):
        self.assertRaises(CMakeSyntaxError, self.listener.syntaxError, None, None, 1, 1, "Test message", None)

    def test_ambiguity(self):
        # Our grammar is now ambiguous, so ensure reporting an ambiguity does not raise an exception
        self.listener.reportAmbiguity(None, None, None, None, None, None, None)

    def test_full_context(self):
        # No-op, we include it for documentation purposes
        # Will be called if an SLL conflict occurs but before a full-context LL resolution is made
        # Not an error, only a notification of a differing method than normal resolution
        self.assertEqual(None, self.listener.reportAttemptingFullContext(None, None, None, None, None, None))

    def test_context_sensitivity(self):
        # May be called in normal operating conditions and does not necessarily represent an error
        # Included for documentation purposes
        # No-op
        self.assertEqual(None, self.listener.reportContextSensitivity(None, None, None, None, None, None))



# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
#
# file: context.py
#
# This file contains paths and other helpful information used throughout the
# test suites of CMinx
#
################################################################################

import os
import sys

# The current working directory
tests_dir = os.path.dirname(__file__)

# The root directory of the project
root_dir = os.path.join(tests_dir, "..")
sys.path.insert(0, os.path.abspath(root_dir))

import cminx

# Here we work out some paths for use throughout the test suite

# The directory containing examples
example_dir = os.path.abspath(os.path.join(tests_dir, "examples"))

# The examples.cmake file in the example_dir
example_cmake = os.path.join(example_dir, "example.cmake")

# The sphinx directory in the example_dir
example_sphinx = os.path.join(example_dir, "sphinx")

# A prefix to be used to test CMinx's prefix option.
prefix = "prefix"

# The directory containing test_samples
test_samples_dir = os.path.abspath(os.path.join(tests_dir, "test_samples"))

# The correct examples.rst file
corr_example_rst = os.path.join(example_sphinx, "source", "example.rst")

# The correct index.rst file
corr_index_rst = os.path.join(test_samples_dir, "corr_rst", "index", "index.rst")

# The correct examples.rst file, with context.prefix as a prefix
corr_example_prefix_rst = os.path.join(example_sphinx, "source", "example_prefix.rst")

# The correct index.rst file, with context.prefix as a prefix
corr_index_prefix_rst = os.path.join(test_samples_dir, "corr_rst", "index", "index_prefix.rst")

corr_example_no_undocumented_diff_header = os.path.join(example_sphinx, "source",
                                                        "example_no_undocumented_diff_header.rst")

corr_example_no_undocumented = os.path.join(example_sphinx, "source", "example_no_undocumented.rst")

corr_index_no_auto_exclude = os.path.join(example_sphinx, "source", "index_no_auto_exclude.rst")

corr_index_exclusion_filters = os.path.join(example_sphinx, "source", "index_exclusion_filters.rst")

# Where the example configs are located
config_dir = os.path.join(example_dir, "configs")

# The config that sets all the input.include_undocumented_* settings to False
no_include_undocumented_config = os.path.join(config_dir, "no_include_undocumented_config.yaml")

# The config that sets auto_exclude_directories_without_cmake to False
no_auto_exclude_config = os.path.join(config_dir, "no_auto_exclude_config.yaml")

# The config that sets the exclusion filters to ignore more_cmake_files/
exclusion_filters_config = os.path.join(config_dir, "exclude_filters_config.yaml")

# The config that excludes example.cmake
exclude_input_file_config = os.path.join(config_dir, "exclude_input_file_config.yaml")



# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Generate Sphinx-compatible RST documentation for CMake files.
Documentation is written in a special form of block comments,
denoted by the starting characters :code: `#[[[` and ending with the standard :code: `#]]`.

:Author: Branden Butler
:License: Apache 2.0
"""

import argparse
import copy
import logging
import logging.config
import os
import re
import sys
import logging.config
from typing import List

import pathspec
from confuse import Configuration
from pkg_resources import get_distribution, DistributionNotFound

from .config import config_template, dict_to_settings, Settings
from .documenter import Documenter
from .rstwriter import RSTWriter

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = "UNKNOWN"

logger: logging.Logger = logging.getLogger(__name__)


def main(args: List[str] = tuple(sys.argv[1:])):
    """
    CMake Documentation Generator program entry point.

    :param args: Array of strings containing program arguments, excluding program name. Same format as sys.argv[1:].
    """

    global logger

    parser = argparse.ArgumentParser(
        description="Automatic documentation generator for CMake files. This program " +
                    "generates Sphinx-compatible RST documents, which are incompatible "
                    "with standard docutils. Config files are searched for according to "
                    "operating-system-dependent directories, such as "
                    "$XDG_CONFIG_HOME/cminx on Linux. Additional config files can be "
                    "specified with the -s option.")
    parser.add_argument(
        "files",
        nargs="+",
        help="CMake file to generate documentation for. If directory, will generate documentation for "
             "all *.cmake files (case-insensitive)")
    parser.add_argument(
        "-o",
        "--output",
        help="Directory to output generated RST to. If not specified will print to standard output. "
             "Output files will have the original filename with the cmake extension replaced by .rst",
        dest="output.directory")
    parser.add_argument(
        "-r",
        "--recursive",
        help="If specified, will generate documentation for all subdirectories of specified directory "
             "recursively. If the prefix is not specified, it will be set to the last element of the "
             "input path.",
        action="store_true",
        default=None,
        dest="input.recursive")
    parser.add_argument(
        "-p",
        "--prefix",
        help="If specified, all output files will have headers generated as if the prefix was the top "
             "level package.",
        dest="rst.prefix")
    parser.add_argument(
        "-s",
        "--settings",
        help="Load settings from the specified YAML file. Parameters specified by "
             "this file will override defaults, and command-line arguments will "
             "override both.")
    parser.add_argument(
        "-e",
        "--exclude",
        help="Exclude some files from documentation. Supports shell-style glob syntax, relative paths "
        "are resolved with respect to the current working directory.",
        dest="input.exclude_filters",
        action="append")
    parser.add_argument(
        "--version",
        action='version',
        version='%(prog)s version ' +
                __version__)

    args = parser.parse_args(args)
    settings = Configuration("cminx", __name__)

    if args.settings is not None:
        # Additional settings file was defined on the command line
        settings.set_file(os.path.abspath(args.settings))

    settings.set_args(args, dots=True)

    output_dir_relative_to_config = False

    if settings["output"]["relative_to_config"].get():
        output_dir_relative_to_config = True

    settings_dict = settings.get(
        config_template(output_dir_relative_to_config))

    settings_obj = dict_to_settings(settings_dict)

    # Concatenate all exclude filters rather than overriding the entire list
    settings_obj.input.exclude_filters = list(
        settings["input"]["exclude_filters"].all_contents())

    # Load Python logging configuration from settings
    logging.config.dictConfig(settings_obj.logging.logger_config)
    logger = logging.getLogger(__name__)

    logger.debug(f"Exclude filters: {settings_obj.input.exclude_filters}")

    if settings_dict["output"]["directory"] is not None:
        output_path = os.path.abspath(settings_dict["output"]["directory"])
        logger.info(f"Writing RST files to {output_path}")

    for input_file in args.files:
        # Process all files specified on command line
        document(input_file, settings_obj)


def document(input_file: str, settings: Settings):
    """
    Handler for documenting CMake files or all files in a directory. Performs
    preprocessing before handing off to document_single_file over all detected
    files. Also generates index.rst files for all directories.

    :param input_file: String locating a file or directory to document.
    :param settings: Object containing all necessary settings that will be passed down to the
    documenter, aggregator, and RST writer.
    """
    output_path: str = settings.output.directory
    recursive = settings.input.recursive
    prefix = settings.rst.prefix

    # We copy because we're going to modify the
    # settings to pass down to lower layers during
    # preprocessing
    new_settings = copy.deepcopy(settings)

    input_path = os.path.abspath(input_file)
    if os.path.isdir(input_path):
        # os.path.join() adds a trailing slash to directories if absent
        input_path = os.path.join(input_path, '')

    spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern,
        settings.input.exclude_filters)

    logger.debug(f"Input path: {input_path}")

    # If the input path matches the exclusion pattern, then ignore
    # this whole path
    if spec.match_file(input_path):
        return

    if not os.path.exists(input_path):
        logger.error(f"File or directory \"{input_path}\" does not exist")
        exit(-1)
    elif os.path.isfile(input_path):
        # If the input was a regular file,
        # don't do anything special and just document the single file
        if output_path is not None:
            os.makedirs(output_path, exist_ok=True)
        document_single_file(input_path, input_path, new_settings)
    elif os.path.isdir(input_path):
        # Input was a directory, so we need to do a lot of preprocessing

        last_dir_element = os.path.basename(os.path.normpath(input_file))
        prefix = prefix if prefix is not None else last_dir_element
        new_settings.rst.prefix = prefix

        # Walk dir and add cmake files to list
        for root, subdirs, filenames in os.walk(
                input_path, topdown=True, followlinks=settings.input.follow_symlinks):

            logger.debug(f"Subdirs: {subdirs}")
            logger.debug(f"Root: {root}")

            # Check our subdirs and see if any match the exclusion filters
            # If they do, remove from the list and os.walk() will ignore them
            for subdir in subdirs:
                # The extra os.path.join() with an empty string ensures the
                # directory has a trailing slash
                if spec.match_file(
                    os.path.join(
                        root,
                        os.path.join(
                            subdir,
                            ""))):
                    subdirs.remove(subdir)

            # Check if any files match the exclusion filters
            # If they do, remove them and the rest of the processing
            # will ignore them
            for file in filenames:
                if spec.match_file(os.path.join(root, file)):
                    filenames.remove(file)

            # Check subdirs and files to make sure .cmake files
            # are present, if not then ignore
            if settings.input.auto_exclude_directories_without_cmake:
                # Make a copy because modifying while iterating results in
                # skipping some entries
                for subdir in copy.copy(subdirs):
                    logger.debug(f"Checking filenames in subdir {subdir}")
                    for filename in os.scandir(os.path.join(root, subdir)):
                        if filename.is_file() and filename.path.endswith(".cmake"):
                            break
                    # If we exited loop normally, i.e. a .cmake file was not
                    # found
                    else:
                        subdirs.remove(subdir)

                # Check if files in current dir contain .cmake
                # If not, ignore this dir and continue walking
                for filename in filenames:
                    if filename.endswith(".cmake"):
                        break
                else:
                    continue

            # Sort filenames and subdirs in alphabetical order
            filenames = sorted(filenames)
            subdirs = sorted(subdirs)

            # If we want to output to an actual file
            # and not stdout
            if output_path is not None:
                path = os.path.join(
                    output_path, os.path.relpath(
                        root, input_path))
                # Make sure we have all the directories created
                os.makedirs(path, exist_ok=True)

                rel_path = os.path.relpath(root, input_path)

                # We need to create an index.rst file for each directory
                index = RSTWriter(rel_path, settings=settings)

                if prefix is not None:
                    # If current file dir is same as root dir, replace "." with
                    # prefix
                    if index.title == settings.rst.module_path_separator:
                        index.title = prefix
                    else:
                        # Add prefix to beginning of header
                        index.title = prefix + settings.rst.module_path_separator + index.title

                toctree = index.directive("toctree")
                toctree.option("maxdepth", 2)

                if recursive:
                    for directory in subdirs:
                        toctree.text(os.path.join(directory, "index.rst"))

                # Filter filenames for cmake files, then add a toctree entry
                for file in [f for f in filenames if f.lower().endswith(".cmake")]:
                    toctree.text('.'.join(file.split('.')[:-1]))


                index.write_to_file(
                    os.path.join(
                        os.path.join(
                            output_path,
                            rel_path),
                        "index.rst"))

            # All preprocessing is complete and we have
            # our index.rst, now just loop over files with the .cmake
            # extension and document them
            for file in filenames:
                if "cmake" == file.split(".")[-1].lower():
                    document_single_file(
                        os.path.join(
                            root,
                            file),
                        input_path,
                        new_settings)

            if not recursive:
                break

    else:
        logger.error("File is a special file (socket, FIFO, device file) and is unsupported")


def document_single_file(file, root, settings: Settings):
    """
    Documents a single file, generates the RST, and places the file in the respective directory if output_dir
    specified.

    :param file: Path to the CMake file to be documented
    :param root: Directory considered to be the root of the source tree. The RST header and output tree will be
    generated from the relative path between file and root
    :param settings: Object containing all necessary settings that will be passed down to the
    documenter, aggregator, and RST writer.
    """

    output_path: str = settings.output.directory
    prefix = settings.rst.prefix
    module_path_separator = settings.rst.module_path_separator

    if os.path.isdir(root):
        # Path to file relative to input_path
        header_name = os.path.relpath(file, root)
    else:
        header_name = file

    if prefix is not None:
        # If current file dir is same as root dir, replace "." with prefix
        if header_name == module_path_separator:
            header_name = prefix
        else:
            # Add prefix to beginning of headers
            header_name = prefix + module_path_separator + header_name

    module_name = header_name

    # Clean header and module names if desired
    # by removing extension

    if not settings.rst.file_extensions_in_titles:
        header_name = re.sub(r"\.cmake$", "", header_name)

    if not settings.rst.file_extensions_in_modules:
        module_name = re.sub(r"\.cmake$", "", module_name)

    # Only log when not writing to stdout
    if output_path is not None:
        logger.info(f"Writing for file {file}")

    auto_documenter = Documenter(file, header_name, module_name, settings)

    output_writer = auto_documenter.process()
    if output_path is not None:
        # Determine where to place generated RST file
        os.makedirs(output_path, exist_ok=True)
        if os.path.isdir(output_path):
            output_filename = os.path.join(
                output_path, ".".join(
                    os.path.basename(file).split(".")[
                    :-1]) + ".rst")
            if os.path.isdir(root):
                # Path to file relative to input_path
                subpath = os.path.relpath(file, root)
                output_filename = os.path.join(
                    output_path, os.path.join(
                        os.path.dirname(subpath), ".".join(
                            os.path.basename(file).split(".")[
                            :-1]) + ".rst"))
            logger.info(f"Writing RST file {output_filename}")
            output_writer.write_to_file(output_filename)
    else:  # Output was not specified so print to screen
        # Use print() for raw output instead of logger
        print(str(output_writer) + "\n")



# Copyright 2022 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses


@dataclasses.dataclass
class CMakeSyntaxException(Exception):
    """
    Denotes a syntax error in the CMake source code.
    This is usually not recoverable and analysis
    of the CMake code is not guaranteed to be consistent,
    so generation of this exception is expected to
    halt the program.
    """

    msg: str
    """A message describing the syntax error."""

    line: int
    """
    The line in the CMake file where the exception was detected.
    Note that this may not be the line where the error is actually located
    in the source, only where the lexer or parser detected an error.
    """



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from cminx.rstwriter import RSTWriter

#First construct the root writer, only argument is the document title
writer = RSTWriter("This is a test")

#Next add any fields needed, such as author, copyright, or license.
writer.field("Author", "Branden Butler")

#Add a paragraph to the root section, after the fields.
writer.text("Text")

#Add a doctest for both showing usage and to ensure future revisions still work the same.
writer.doctest(r"print('test')", r"test")

#General admonition, we want people to know about John.
d = writer.directive("admonition", "John")

#John's social network
d.text("Social network")

#For other directives you may add options, such as the one below for toctree
#d.option('maxdepth', 2)
#d.text('file/to/include')

#John wants an image under his admonition
img = d.directive("image", "image.png")

#Images can also have options, here we change the image width and height, stretching it
img.option("width", 500)
img.option("height", 100)

#Adding a subsection to the root (appears after John's admonition, social network, and image)
s = writer.section('A subsection')
s.text("Subsection's content")

#Adding lists. These use varargs so if you have a prebuilt list already you'll want to splat it out
s.bulleted_list("Item 1", "Item 2", "Item 3")
s.enumerated_list("Item 1", "Item 2", "Item 3")

#Make a table of values
tab = [
[0, 1, 2, 3],
[4, 5, 6, 7],
[8, 9, 10, 11]
]
s.simple_table(tab, column_headings=["Column 1", "Column 2", "Column 3", "Column 4"])

#Automatically rebuild header whenever title changes
writer.title = "This title is different"

#Comment below and 'with' statement are equivalent, shows that write_to_file can handle both strings and file-like objects
#writer.write_to_file("file.rst")
with open("file.rst", "w") as f:
        writer.write_to_file(f)




#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import textwrap
import unittest

import pytest

import context
from antlr4 import *

from cminx import Settings
from cminx.config import InputSettings
from cminx.exceptions import CMakeSyntaxException
from cminx.parser import ParserErrorListener
from cminx.parser.CMakeLexer import CMakeLexer
from cminx.parser.CMakeParser import CMakeParser
from cminx.aggregator import DocumentationAggregator
from cminx.documentation_types import FunctionDocumentation, MacroDocumentation, VariableDocumentation, \
    GenericCommandDocumentation, ClassDocumentation, VarType, DocumentationType, MethodDocumentation


class TestAggregator(unittest.TestCase):

    def setUp(self):
        self.filename = context.example_cmake
        # We need a string stream of some kind, FileStream is easiest
        self.input_stream = FileStream(self.filename)
        self.reset()

    def reset(self, settings=Settings()):
        # Convert those strings into tokens and build a stream from those
        self.lexer = CMakeLexer(self.input_stream)
        self.stream = CommonTokenStream(self.lexer)

        # We now have a stream of CommonToken instead of strings, parsers require this type of stream
        self.parser = CMakeParser(self.stream)
        self.parser.addErrorListener(ParserErrorListener())
        self.tree = self.parser.cmake_file()

        # Hard part is done, we now have a fully useable parse tree, now we just need to walk it
        self.aggregator = DocumentationAggregator(settings)
        self.walker = ParseTreeWalker()
        self.walker.walk(self.aggregator, self.tree)

    def test_aggregated_docs(self):

        # All of the documented commands are now stored in aggregator.documented,
        # each element is a namedtuple repesenting the type of documentation it is.
        # So far we can document functions, macros, and variables (only strings and lists built using set)
        for doced_command in self.aggregator.documented:
            self.assertNotEqual(doced_command, None)
            self.assertIsInstance(doced_command, DocumentationType, "Unknown documentation type")

    def test_doccomment_function_leading_space(self):
        docstring = "This is a function"
        function_name = "TEST_FUNC"
        params = ["param1", "param2"]

        self.input_stream = InputStream(f'''
#[[[
 {docstring}
#]]
function({function_name} {params[0]} {params[1]})

endfunction()
        ''')
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), FunctionDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), function_name,
                         "Incorrect function_name extracted")
        self.assertListEqual([param.strip() for param in self.aggregator.documented[0].params], params,
                             "Incorrect params extracted")

    def test_doccomment_macro_leading_space(self):
        docstring = "This is a macro"
        macro_name = "testmacro"
        params = ["param1", "param2"]

        self.input_stream = InputStream(f'''
#[[[
 {docstring}
#]]
macro({macro_name} {params[0]} {params[1]})

endmacro()
        ''')
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), MacroDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), macro_name, "Incorrect macro_name extracted")
        self.assertListEqual([param.strip() for param in self.aggregator.documented[0].params], params,
                             "Incorrect params extracted")

    def test_doccomment_stringvar_leading_space(self):
        docstring = "This is a string variable"
        var_name = "TEST_VAR"
        val = "This is a value"

        self.input_stream = InputStream(textwrap.dedent(f'''
                                #[[[
                                 {docstring}
                                #]]
                                set({var_name} "{val}")

                            '''))
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), VariableDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), var_name, "Incorrect function_name extracted")
        self.assertEqual(self.aggregator.documented[0].type, VarType.STRING)
        self.assertEqual(self.aggregator.documented[0].value.strip(), val.strip(), "Incorrect value extracted")

    def test_doccomment_listvar_leading_space(self):
        docstring = "This is a list variable"
        var_name = "listvar"
        params = ["param1", "param2"]

        self.input_stream = InputStream(textwrap.dedent(f'''
                                #[[[
                                {docstring}
                                #]]
                                set({var_name} {params[0]} {params[1]})
                            '''))
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), VariableDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), var_name, "Incorrect var_name extracted")
        self.assertEqual(self.aggregator.documented[0].type, VarType.LIST)

        self.assertEqual(self.aggregator.documented[0].value, " ".join(params),
                         "Incorrect list elements extracted")

    def test_unset(self):
        docstring = "Unsetting a variable"
        var_name = "myvar"

        self.input_stream = InputStream(textwrap.dedent(f'''
                                #[[[
                                {docstring}
                                #]]
                                set({var_name})
                            '''))
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), VariableDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), var_name, "Incorrect var_name extracted")
        self.assertEqual(self.aggregator.documented[0].type, VarType.UNSET)

    def test_cpp_class_no_superclass_no_inner(self):
        self.test_cpp_class_multi_superclass_no_inner([])

    def test_cpp_class_multi_superclass_multi_members(self,
                                                      superclasses=("SuperClassA", "SuperClassB", "SuperClassC"),
                                                      attributes=("attr1", "attr2"),
                                                      methods=("method1", "method2"),
                                                      inner_classes=("Inner1", "Inner2")):
        class_docstring = "This is a class"
        inner_class_docstring = "#[[[\n# This is an inner class\n#]]"
        method_docstring = "#[[[\n# This is a method\n#]]"
        attribute_docstring = "#[[[\n# This is an attribute\n#]]"
        class_name = "MyClass"
        inner_class_definitions = '\n'.join(
            [
                textwrap.dedent(f"""\
                    {inner_class_docstring}
                    cpp_class({inner_class_name})
                    cpp_end_class()
                """)
                for inner_class_name in inner_classes
            ]
        )
        # Curly brackets in f strings are escaped with two brackets. So we end up with "${<value of method_name>}"
        method_definitions = '\n'.join(
            [
                textwrap.dedent(f"""\
                    {method_docstring}
                    cpp_member({method_name} {class_name})
                    function(${{{method_name}}})
                    endfunction()
                """)
                for method_name in methods
            ]
        )
        attribute_definitions = '\n'.join(
            [f'{attribute_docstring}\ncpp_attr({class_name} {attr_name})' for attr_name in attributes])
        self.input_stream = InputStream(textwrap.dedent(f'''
                                #[[[
                                # {class_docstring}
                                #]]
                                cpp_class({class_name} {' '.join(superclasses)})

                                    {attribute_definitions}

                                    {method_definitions}

                                    {inner_class_definitions}

                                cpp_end_class()
                            '''))
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1 + len(inner_classes),
                         "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), ClassDocumentation, "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), class_docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), class_name, "Incorrect class name extracted")
        self.assertEqual(len(self.aggregator.documented[0].superclasses), len(superclasses),
                         "Superclasses incorrectly found")
        for i in range(len(superclasses)):
            self.assertEqual(self.aggregator.documented[0].superclasses[i].strip(), superclasses[i].strip(),
                             "Superclass name not preserved")

        self.assertEqual(len(self.aggregator.documented[0].inner_classes), len(inner_classes),
                         "Inner classes incorrectly found")
        self.assertEqual(len(self.aggregator.documented[0].members), len(methods), "Members incorrectly found")
        self.assertEqual(len(self.aggregator.documented[0].attributes), len(attributes), "Attributes incorrectly found")

    def test_cpp_class_multi_superclass_no_inner(self, superclasses=("SuperClassA", "SuperClassB", "SuperClassC")):
        self.test_cpp_class_multi_superclass_multi_members(superclasses=superclasses, attributes=[], methods=[])

    def test_cpp_class_one_superclasses_no_inner(self):
        self.test_cpp_class_multi_superclass_no_inner(["SuperClass"])

    def test_unknown_documented_command(self):
        docstring = 'This is documentation for an unknown command'
        params = ["FATAL_ERROR", "\"This is a test\""]
        command_name = "message"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(len(self.aggregator.documented), 1, "Different number of documented commands than expected")
        self.assertEqual(type(self.aggregator.documented[0]), GenericCommandDocumentation,
                         "Unexpected documentation type")
        self.assertEqual(self.aggregator.documented[0].doc.strip(), docstring, "Incorrect docstring extracted")
        self.assertEqual(self.aggregator.documented[0].name.strip(), command_name, "Incorrect command name extracted")
        for i in range(0, len(params)):
            p = self.aggregator.documented[0].params[i]
            self.assertEqual(params[i], p, "Incorrect command parameters. Expected {params[i]}, got {p}")

    def test_incorrect_function_params(self):
        docstring = "This is documentation for an incorrect function() call"
        params = []
        command_name = "function"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        with pytest.raises(CMakeSyntaxException):
            self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_macro_params(self):
        docstring = "This is documentation for an incorrect macro() call"
        params = []
        command_name = "macro"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        with pytest.raises(CMakeSyntaxException):
            self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_ct_add_test_params(self):
        docstring = "This is documentation for an incorrect ct_add_test() call"
        # ct_add_test() requires an additional param after NAME
        params = ["NAME"]
        command_name = "ct_add_test"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_invalid_ct_add_test_params(self):
        docstring = "This is documentation for an incorrect ct_add_section() call"
        # ct_add_test() requires an additional param after NAME, not before
        params = ["test_name", "NAME"]
        command_name = "ct_add_test"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_ct_add_section_params(self):
        docstring = "This is documentation for an incorrect ct_add_section() call"
        # ct_add_section() requires an additional param after NAME
        params = ["NAME"]
        command_name = "ct_add_section"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_invalid_ct_add_section_params(self):
        docstring = "This is documentation for an incorrect ct_add_section() call"
        # ct_add_section() requires an additional param after NAME, not before
        params = ["section_name", "NAME"]
        command_name = "ct_add_section"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_invalid_ctest_add_test_params(self):
        docstring = "This is documentation for an incorrect ct_add_section() call"
        # add_test() requires two params minimum
        params = ["NAME"]
        command_name = "add_test"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_ctest_add_test_params(self):
        docstring = "This is documentation for an incorrect ct_add_section() call"
        # add_test() requires an additional param after NAME, not before
        params = ["blah", "NAME"]
        command_name = "add_test"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_set_params(self):
        docstring = "This is documentation for an incorrect set() call"
        params = []
        command_name = "set"
        command = f'{command_name}({" ".join(params)})'
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_cpp_class_params(self):
        params = []
        command_name = "cpp_class"
        command = f'{command_name}({" ".join(params)})'
        docstring = f"This is documentation for an incorrect {command_name}() call"
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_cpp_member_params(self):
        params = []
        command_name = "cpp_member"
        command = f'{command_name}({" ".join(params)})'
        docstring = f"This is documentation for an incorrect {command_name}() call"
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_cpp_attribute_params(self):
        params = []
        command_name = "cpp_attr"
        command = f'{command_name}({" ".join(params)})'
        docstring = f"This is documentation for an incorrect {command_name}() call"
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_incorrect_option_params(self):
        params = []
        command_name = "option"
        command = f'{command_name}({" ".join(params)})'
        docstring = f"This is documentation for an incorrect {command_name}() call"
        self.input_stream = InputStream(f'#[[[\n{docstring}\n#]]\n{command}')
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"Incorrect {command_name}() call was still added to documented list: {self.aggregator.documented}")

    def test_cpp_attr_outside_class(self):
        self.input_stream = InputStream("#[[[\n# cpp_attr() outside class\n#]]\ncpp_attr()")
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"cpp_attr() call outside class was still added to documented list: {self.aggregator.documented}")

    def test_cpp_member_outside_class(self):
        self.input_stream = InputStream("#[[[\n# cpp_member() outside class\n#]]\ncpp_member()")
        self.reset()
        self.assertEqual(0, len(self.aggregator.documented),
                         f"cpp_member() call outside class was still added to documented list: {self.aggregator.documented}")

    def test_invalid_syntax(self):
        self.input_stream = InputStream("#[[[\n#invalid syntax\n#]]\nfunction()\nend_function()")
        with pytest.raises(CMakeSyntaxException):
            self.reset()

    def test_function_param_regex(self):
        func_name = "test_func"
        stripped_param_names = ["param1", "param2_with_underscores"]
        param_names = [f"_tf_{name}" for name in stripped_param_names]

        docstring = "This is a function that has its param name regex-stripped"

        regex = "^_[a-zA-Z]*_"

        self.input_stream = InputStream(textwrap.dedent(f"""
                    #[[[
                    # {docstring}
                    #]]
                    function({func_name} {' '.join(param_names)})
                    endfunction()
                """))

        input_settings = InputSettings(function_parameter_name_strip_regex=regex)
        self.reset(settings=Settings(input=input_settings))
        self.assertIsInstance(self.aggregator.documented[0], FunctionDocumentation)
        self.assertEqual(func_name, self.aggregator.documented[0].name)
        for i in range(0, len(stripped_param_names)):
            self.assertEqual(
                stripped_param_names[i], self.aggregator.documented[0].params[i],
                "Parameter name was not stripped correctly"
            )

    def test_macro_param_regex(self):
        func_name = "test_macro"
        stripped_param_names = ["param1", "param2_with_underscores"]
        param_names = [f"_tm_{name}" for name in stripped_param_names]

        docstring = "This is a macro that has its param name regex-stripped"

        regex = "^_[a-zA-Z]*_"

        self.input_stream = InputStream(textwrap.dedent(f"""
            #[[[
            # {docstring}
            #]]
            macro({func_name} {' '.join(param_names)})
            endmacro()
        """))

        input_settings = InputSettings(macro_parameter_name_strip_regex=regex)
        self.reset(settings=Settings(input=input_settings))
        self.assertIsInstance(self.aggregator.documented[0], MacroDocumentation)
        self.assertEqual(func_name, self.aggregator.documented[0].name)
        for i in range(0, len(stripped_param_names)):
            self.assertEqual(
                stripped_param_names[i], self.aggregator.documented[0].params[i],
                "Parameter name was not stripped correctly"
            )

    def test_method_param_regex(self):
        func_name = "test_method"
        stripped_param_names = ["param1", "param2_with_underscores"]
        param_names = [f"_tm_{name}" for name in stripped_param_names]
        param_types = ["desc", "str"]

        docstring = "This is a method that has its param name regex-stripped"

        regex = "^_[a-zA-Z]*_"

        self.input_stream = InputStream(textwrap.dedent(f"""
                    cpp_class(TestClass)
                        #[[[
                        # {docstring}
                        #]]
                        cpp_member({func_name} TestClass {' '.join(param_types)})
                        function("${{{func_name}}}" self {' '.join(param_names)})
                        endfunction()
                    cpp_end_class()
                """))

        input_settings = InputSettings(member_parameter_name_strip_regex=regex)
        self.reset(settings=Settings(input=input_settings))
        self.assertIsInstance(self.aggregator.documented[0], ClassDocumentation)
        self.assertIsInstance(self.aggregator.documented[0].members[0], MethodDocumentation)
        self.assertEqual(func_name, self.aggregator.documented[0].members[0].name)
        for i in range(0, len(stripped_param_names)):
            self.assertEqual(
                stripped_param_names[i], self.aggregator.documented[0].members[0].params[i],
                "Parameter name was not stripped correctly"
            )


if __name__ == '__main__':
    unittest.main()




# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
If CMinx is built with CMake, we use PyInstaller to create an executable.
PyInstaller requires us to import CMinx so it can discover the contents of
CMinx. This file imports CMinx and calls its entry point so PyInstaller can
do its thing.

:Author: CMakePP
:License: Apache 2.0
"""

from cminx import *

if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from antlr4 import *
from cminx.parser.CMakeLexer import CMakeLexer
from cminx.parser.CMakeParser import CMakeParser
from aggregator import DocumentationAggregator


def main(argv):
    #We need a string stream of some kind, FileStream is easiest
    input_stream = FileStream(argv[1])

    #Convert those strings into tokens and build a stream from those
    lexer = CMakeLexer(input_stream)
    stream = CommonTokenStream(lexer)

    #We now have a stream of CommonToken instead of strings, parsers require this type of stream
    parser = CMakeParser(stream)
    tree = parser.cmake_file()

    #Hard part is done, we now have a fully useable parse tree, now we just need to walk it
    aggregator = DocumentationAggregator()
    walker = ParseTreeWalker()
    walker.walk(aggregator, tree)

    #All of the documented commands are now stored in aggregator.documented,
    #each element is a namedtuple repesenting the type of documentation it is.
    #So far we can document functions, macros, and variables (only strings and lists built using set)
    print(aggregator.documented)

if __name__ == '__main__':
    #Usage: aggregator_example.py <file>.cmake
    main(sys.argv)



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

from cminx.rstwriter import RSTWriter
from .rst_validator import RSTValidator


class TestRSTWriter(unittest.TestCase):
    def setUp(self):
        self.writer = RSTWriter("Testing RSTWriter")
        self.validator = RSTValidator(self.writer, werror=True)  # Treat all warnings as validation failure
        self.file = "writer_test.rst"

    def tearDown(self):
        try:
            os.remove(self.file)
        except FileNotFoundError:
            pass  # Test just didn't write to the file

    def test_new_title(self):
        new_title = "This is a new title"
        self.writer.title = new_title
        self.assertTrue(new_title in str(self.writer), "Title was not updated")
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_paragraph(self):
        text = "This is a paragraph"
        self.writer.text(text)
        self.assertTrue(self.validator.validate(), self.validator.failures)
        self.assertTrue(text in self.writer.to_text(), "Paragraph was not included")

    def test_admonition(self):
        admonition_type = "DANGER"
        text = "Beware of the killer rabbits"
        d = self.writer.directive(admonition_type)
        d.text(text)
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_image(self):
        file = "image.png"
        img = self.writer.directive("image", file)
        img.option("width", 500)
        img.option("height", 100)
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_doctest(self):
        test_line = 'print("test")'
        expected_output = 'test'
        self.writer.doctest(test_line, expected_output)
        self.assertTrue(test_line in self.writer.to_text(), "DocTest test line was not included")
        self.assertTrue(expected_output in self.writer.to_text(), "DocTest expected output was not included")
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_lists(self):
        items = ["Test 1", 2, 16.4, True]
        self.writer.bulleted_list(*items)
        self.assertTrue(self.validator.validate(), self.validator.failures)
        self.writer.clear()
        self.writer.enumerated_list(*items)
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_field(self):
        name = "TestField"
        text = "This is a field"
        self.writer.field(name, text)
        self.assertTrue(self.validator.validate(), self.validator.failures)
        self.assertTrue(name in self.writer.to_text(), "Field name was not included")
        self.assertTrue(text in self.writer.to_text(), "Field text was not included")

    def test_table(self):
        tab = [
            [0, 1, 2, 3],
            [4, 5, 6, 7],
            [8, 9, 10, 11]
        ]
        headings = ["Column 1", "Column 2", "Column 3", "Column 4"]
        self.writer.simple_table(tab, column_headings=headings)
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_section(self):
        text = "This is inside a section"
        sect = self.writer.section("Section")
        sect.text(text)
        self.assertTrue(text in str(self.writer), f"Subsection text was not included")
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_directive(self):
        dir_type = "attention"

        opts = {"name": "attention"}
        dir_text = "This is an attention admonition"

        directive = self.writer.directive(dir_type)
        for opt_type, opt_val in opts.items():
            directive.option(opt_type, opt_val)

        directive.text(dir_text)

        self.assertTrue(dir_type in str(self.writer), f"Directive {dir_type} was not included")
        self.assertTrue(self.validator.validate(), self.validator.failures)

    def test_write(self):
        self.writer.write_to_file(self.file)
        with open(self.file, "w") as f:
            self.writer.write_to_file(f)

        with self.assertRaises(ValueError):
            self.writer.write_to_file("")

        with self.assertRaises(ValueError):
            self.writer.write_to_file(None)

        with self.assertRaises(TypeError):
            self.writer.write_to_file(12)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import docutils.frontend
import docutils.parsers.rst
import docutils.utils
from docutils.nodes import paragraph, section, bullet_list, enumerated_list, table, system_message, doctest_block, \
    Admonition, field_list, image, tbody

from cminx.rstwriter import RSTWriter, Paragraph, RSTList, Field, DocTest, SimpleTable, Directive


class ValidationError(RuntimeError):
    """
    Raised when something seriously wrong happens when validating RST structures.
    Situations in which this exception will be raised include:

    1. No processor mapping found for a particular RST element
    2. A 'severe' level message occurs somewhere in the RST document tree
    3. The parsed section and the RSTWriter section have two different lengths
    """
    pass


class RSTValidator:
    """
    Validates an RSTWriter document by feeding it through DocUtils and making sure everything comes out the same.
    This has the disadvantage of being unable to validate structures unsupported by DocUtils, such as toctree,
    but it works for testing to make sure RSTWriter outputs correct syntax.
    Currently supported validation targets:
    1. Sections and headings
    2. Paragraphs
    3. Both enumerated and bulleted lists
    4. Simple tables
    5. Doctests
    6. Fields
    7. Admonitions, such as DANGER. Other directive types not supported

    """

    def __init__(self, writer: RSTWriter, werror=False, doc=None):
        """
        :param writer: RSTWriter that will be validated.

        :param werror: Whether DocUtils warnings should be treated as a failed validation.

        :param doc: Optional DocUtils document to validate against. If None, a document will be generated from the output of writer
        """
        self.mapping = {
            paragraph: self.process_paragraph,
            section: self.process_sect,
            bullet_list: self.process_list,
            enumerated_list: self.process_list,
            table: self.process_table,
            system_message: self.process_message,
            doctest_block: self.process_doctest,
            Admonition: self.process_admonition,
            field_list: self.process_field_list,
            image: self.process_image
        }
        self.writer = writer
        self.parser = docutils.parsers.rst.Parser()
        self.doc = doc
        self.validated = False
        self.werror = werror
        self.failures = []

    def validate(self) -> bool:
        """
        Parse the RSTWriter text and process the document tree. Previous validated result will be wiped.
        This method may take awhile to run if validating larger documents, so cache the result if necessary.

        :return: Whether the supplied RSTWriter succeeded in being validated.

        :raises ValidationError: If validation encounters an unrecoverable or severe problem.
        """

        self.validated = True  # Assume everything works until something fails down the line
        self.failures = []
        if self.doc is None:
            self.doc = RSTValidator.parse_rst(self.parser,
                                              self.writer.to_text())
        self.process_element(self.doc[0], self.writer)
        for msg in self.doc.parse_messages:
            self.process_message(msg)
        return self.validated

    def fail(self, msg):
        """
        Record a validation failure

        :param msg: A string saying what happened that caused a failed validation
        """

        self.failures.append(msg)
        self.validated = False

    @staticmethod
    def parse_rst(parser, text: str) -> docutils.nodes.document:
        """
        Parse the RST text into a DocUtils document, using the RST parser

        :param text: String containing RST document to be parsed
        """

        components = (docutils.parsers.rst.Parser,)
        settings = docutils.frontend.OptionParser(
            components=components).get_default_values()
        document = docutils.utils.new_document('<rst-doc>', settings=settings)
        parser.parse(text, document)
        return document

    def process_message(self, msg: system_message):
        """
         Processes all DocUtils messages found in the document tree, as well as those found in the parse_messages list.
         If the level is error and higher, or warning and higher if werror is True, then the validation status is set to False.

         :param msg: DocUtils message node to be processed

         :raises ValidationError: If message's level is severe
         """

        level = msg['level']
        if level >= (2 if self.werror else
        3):  # Warn or higher if werror, else only error or higher
            self.fail(msg)
        if level >= 3:  # Severe, we need to raise a stink about this
            raise ValidationError(msg)

    def process_element(self, element, to_validate):
        """
        All elements in the document first pass through here.
        The specific processor is found by matching the list of superclasses
        against RSTValidator.mapping in order of method resolution.

        :param element: DocUtils document element, the processor used is found by matching element's class and superclasses.

        :param to_validate: RSTWriter element being validated, should match the DocUtils equivalent

        :raises ValidationError: If the processor used raises it or a processor could not be found.
        """
        for t in type(element).mro(
        ):  # Look up the chain of superclasses until we find a mapping, in method resolution order
            if t in self.mapping:
                self.mapping[t](element, to_validate)
                break
        else:  # For-else clause, executes if break is not called
            raise ValidationError(
                f"No processor mapping found for {type(element)}")

    def process_paragraph(self, para: paragraph, to_validate: Paragraph):
        """
        Processor for the paragraph element.

        :param para: DocUtils paragraph element to validate against.

        :param to_validate: RSTWriter Paragraph element to be validated.
        """
        if para.astext() != to_validate.text:
            self.fail(
                f"Parsed paragraph {para.astext()} != original paragraph {to_validate.text}"
            )

    def process_field_list(self, fields: field_list, to_validate: Field):
        """
        Processor for field lists

        :param fields: DocUtils field_list element to validate against

        :param to_validate): RSTWriter Field element to validate
        """

        name = fields[0][0].astext()
        text = fields[0][1].astext()
        if name != to_validate.field_name:
            self.fail(
                f"Parsed field name '{name}' != original field name '{to_validate.field_name}'"
            )

        if text != to_validate.field_text:
            self.fail(
                f"Parsed field text {text} != original field text '{to_validate.field_text}'"
            )

    def process_list(self, rst_list, to_validate: RSTList):
        """
        Processor for list elements, both bulleted and enumerated.

        :param rst_list: DocUtils list element to validate against.

        :param to_validate: RSTWriter List element to validate.
        """
        for i in range(0, len(rst_list)):
            if rst_list[i].astext() != str(to_validate.items[i]):
                self.fail("Parsed list items do now match original list items")

    def process_admonition(self, ad: Admonition, to_validate: Directive):
        """
        Processor for admonitions, handles all types including general, warning, danger, etc.

        :param ad: DocUtils Admonition element, may be of any subclass

        :param to_validate: RSTWriter Directive element to validate. RSTWriter makes no distinction against types of directives.
        """

        if ad[0].astext() != to_validate.document[1].text:
            self.fail(
                f"Parsed admonition text ({ad[0].astext()}) does not match original admonition text ({to_validate.document[1].text}"
            )

    def process_image(self, img: image, to_validate: Directive):
        """
        Processor for image directives. DocUtils does not have a Directive
        superclass, so even though this method could process all directives
        there is no way to set this as a processor for all of them.

        :param img: DocUtils image element to validate against

        :param to_validate: RSTWriter Directive element to validate. RSTWriter makes no distinction against types of directives.

        :raises ValidationError: If there is an option present in to_validate but not in img
        """
        for option in to_validate.options:
            if option.name in img:
                if str(option.value) != str(img[option.name]):
                    self.fail(
                        "Parsed option values do not match original option values"
                    )
            else:
                raise ValidationError(
                    f"Option {option.name} not in parsed document")

    def process_doctest(self, doctest: doctest_block, to_validate: DocTest):
        """
        Processor for DocTest elements. DocUtils does not store the test_line
        and expected_output separately, so string processing is needed. The triple chevrons
        are removed via a str.lstrip() call, the expected_output is unaltered.

        :param doctest: DocUtils doctest_block element to validate against.

        :param to_validate: RSTWriter DocTest element to validate.
        """
        parts = doctest[0].split('\n')
        test_line = parts[0].lstrip("> ")
        expected_output = "\n".join(parts[1:])
        if test_line != to_validate.test_line or expected_output != to_validate.expected_output:
            self.fail(
                "Parsed test line or expected output do not match either of the original values"
            )

    def process_table(self, t: table, to_validate: SimpleTable):
        """
        Processor for Table elements. DocUtils makes no distinction between simple
        tables and grid tables, this processor can only handle simple tables.

        :param t: DocUtils table element to validate against.

        :param to_validate: RSTWriter SimpleTable element to validate.
        """

        grouping = t[0]
        body_index = grouping.first_child_matching_class(tbody)
        body = grouping[body_index]
        for i in range(0, len(body)):
            for j in range(0, len(body[i])):
                if body[i][j].astext() != str(to_validate.table[i][j]):
                    self.fail(
                        "Parsed table values do not match original table values"
                    )

    def process_sect(self, sect: section, to_validate: RSTWriter):
        """
        Processor for Section elements. Filters out system_message elements and dispatches them
        to their processor before looping over all subelements in the section.

        :param sect: DocUtils section element containing subelements and possibly system messages, each element will be processed separately.

        :param to_validate: RSTWriter document list to validate.

        :raises ValidationError: If any subelements do or the two document trees do not have the same length.
        """
        sect_title = sect[0]
        if sect_title.astext() != to_validate.document[0].title:
            self.fail(
                f"Parsed section title '{sect_title.astext()}' does not match original section title '{to_validate.document[0].title}'"
            )

        section_cleaned = list(
            filter(self.is_not_message, sect)
        )  # Remove all messages and process them before processing rest of the document
        if len(section_cleaned) != len(
                to_validate.document
        ):  # Our two documents are different lengths
            raise ValidationError([
                "Parsed section and RSTWriter section are two different lengths",
                str(section_cleaned),
                str(to_validate.document)
            ])

        for i in range(
                1, len(section_cleaned)
        ):  # Section and to_validate should now have the same number of elements
            self.process_element(section_cleaned[i], to_validate.document[i])

    def is_not_message(self, e) -> bool:
        """
        Determines if e is a message and if so, dispatches it to its processor.

        :param e: Element to check

        :return: True if element is a message, False otherwise
        """
        if system_message in type(e).mro():
            self.process_message(e)
            return False
        else:  # Included for readability
            return True



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import os



def main():
    loader = unittest.TestLoader()
    cwd = os.path.abspath(os.path.dirname(__file__))
    suites = loader.discover(cwd, "test_*.py")
    all_tests = unittest.TestSuite(suites)
    runner = unittest.TextTestRunner()
    result = runner.run(all_tests)

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import context
import unittest
import os
import shutil
import helpers


class TestSamples(unittest.TestCase):

    def setUp(self):
        self.cwd = os.path.abspath(os.path.dirname(__file__))
        self.output_dir = os.path.join(self.cwd, "output")
        self.input_dir = context.test_samples_dir

        # Directory containing the correct answers
        self.corr_dir = os.path.join(self.cwd, "corr_rst")

    def tearDown(self):
        try:
            shutil.rmtree(self.output_dir)
            pass
        except FileNotFoundError:
            pass

    def test_samples(self):
        args = ["-r", "-o", self.output_dir, self.input_dir]
        helpers.quiet_cminx(args)
        for f in os.listdir(self.cwd):
            if ".cmake" not in f:
                continue
            prefix = f.split(".cmake")[0]
            file_name = prefix + ".rst"
            generated_path = os.path.join(self.output_dir, file_name)
            corr_path = os.path.join(self.corr_dir, file_name)
            with open(corr_path, 'r') as corr_file, open(generated_path, 'r') as generated_file:
                self.assertEqual(corr_file.read(), generated_file.read())



# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from antlr4 import *
from antlr4.error.Errors import InputMismatchException

from cminx.parser import ParserErrorListener, CMakeSyntaxError
from cminx.parser.CMakeLexer import CMakeLexer
from cminx.parser.CMakeListener import CMakeListener
from cminx.parser.CMakeParser import CMakeParser


class MockListener(CMakeListener):
    def __init__(self):
        self.elements = []
        """List of parsed element contexts. such as commands and documented commands"""

    def enterDocumented_command(self, ctx: CMakeParser.Documented_commandContext):
        self.elements.append(ctx)

    def enterCommand_invocation(self, ctx: CMakeParser.Command_invocationContext):
        self.elements.append(ctx)


class TestParser(unittest.TestCase):
    def setUp(self):
        self.input_stream = InputStream("")
        self.reset()

    def reset(self):
        # Convert those strings into tokens and build a stream from those
        self.lexer = CMakeLexer(self.input_stream)
        self.stream = CommonTokenStream(self.lexer)

        # We now have a stream of CommonToken instead of strings, parsers require this type of stream
        self.parser = CMakeParser(self.stream)
        self.parser.removeErrorListeners()
        self.parser.addErrorListener(ParserErrorListener())
        self.tree = self.parser.cmake_file()

        # Hard part is done, we now have a fully useable parse tree, now we just need to walk it
        self.aggregator = MockListener()
        self.walker = ParseTreeWalker()
        self.walker.walk(self.aggregator, self.tree)

    def test_empty(self):
        self.reset()
        self.assertEqual(self.aggregator.elements, [])

    def test_function_single_arg(self):
        self.input_stream = InputStream('''
function(MyFunction arg1)\n    message("${arg1}")\nendfunction()
        ''')
        self.reset()
        self.assertEqual(len(self.aggregator.elements), 3, "Different number of extracted elements")
        for element in self.aggregator.elements:
            self.assertEqual(type(element), CMakeParser.Command_invocationContext)
        for i in range(0, len(self.aggregator.elements)):
            element = self.aggregator.elements[i]
            if i == 0:
                # function() command
                self.assertEqual(element.Identifier().getText().lower(), "function")
                self.assertEqual(element.single_argument()[0].Identifier().getText(), "MyFunction")
                params = [param.Identifier().getText() for param in
                          element.single_argument()[1:]]  # Extract declared function parameters
                self.assertListEqual(params, ["arg1"])
            elif i == 1:
                # message() command
                self.assertEqual(element.Identifier().getText().lower(), "message")
                params = [param.Quoted_argument().getText() for param in
                          element.single_argument()]  # Extract message() params
                self.assertListEqual(params, ['"${arg1}"'])
            elif i == 2:
                # endfunction() command
                self.assertEqual(element.Identifier().getText().lower(), "endfunction")
                params = element.single_argument()
                self.assertEqual(params, [])

    def test_function_multi_arg(self):
        self.input_stream = InputStream('''
function(MyFunction arg1 arg2)\n    message("${arg1}" "${arg2}")\nendfunction()
        ''')
        self.reset()
        self.assertEqual(len(self.aggregator.elements), 3, "Different number of extracted elements")
        for element in self.aggregator.elements:
            self.assertEqual(type(element), CMakeParser.Command_invocationContext)
        for i in range(0, len(self.aggregator.elements)):
            element = self.aggregator.elements[i]
            if i == 0:
                # function() command
                self.assertEqual(element.Identifier().getText().lower(), "function")
                self.assertEqual(element.single_argument()[0].Identifier().getText(), "MyFunction")
                params = [param.Identifier().getText() for param in
                          element.single_argument()[1:]]  # Extract declared function parameters
                self.assertListEqual(params, ["arg1", "arg2"])
            elif i == 1:
                # message() command
                self.assertEqual(element.Identifier().getText().lower(), "message")
                params = [param.Quoted_argument().getText() for param in
                          element.single_argument()]  # Extract message() params
                self.assertListEqual(params, ['"${arg1}"', '"${arg2}"'])
            elif i == 2:
                # endfunction() command
                self.assertEqual(element.Identifier().getText().lower(), "endfunction")
                params = element.single_argument()
                self.assertEqual(params, [])

    def test_no_ending_paren(self):
        with self.assertRaises(CMakeSyntaxError):
            self.input_stream = InputStream('''
function(TEST
            ''')
            self.reset()

    def test_mismatch(self):
        with self.assertRaises(InputMismatchException):
            self.input_stream = InputStream('''
()
                ''')
            self.reset()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

import context

from cminx.documenter import Documenter
from cminx.documentation_types import FunctionDocumentation, MacroDocumentation, VariableDocumentation, \
    GenericCommandDocumentation, ClassDocumentation, TestDocumentation, SectionDocumentation, CTestDocumentation, \
    ModuleDocumentation
from cminx.rstwriter import Directive, RSTWriter


class TestDocumenter(unittest.TestCase):

    def setUp(self):
        self.filename = context.example_cmake
        self.reset()

    def reset(self):
        self.documenter = Documenter(self.filename, self.filename)  # File is parsed on __init__

    def test_process(self):
        self.documenter.process()  # Convert all documentation into RST
        self.assertEqual(len(self.documenter.aggregator.documented), len(self.documenter.writer.document) - 1,
                         "Generated RST has different length from input documentation")  # RSTWriter adds one element
        # for document heading, documenter adds another for module definition
        for i in range(0, len(self.documenter.aggregator.documented)):
            doc = self.documenter.aggregator.documented[i]
            element = self.documenter.writer.document[i + 1]
            if isinstance(doc, ModuleDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for module doc")
            elif isinstance(doc, FunctionDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for function")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for function")
            elif isinstance(doc, MacroDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for macro")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for macro")
                self.assertIsInstance(element.document[1], Directive)
                self.assertEqual("note", element.document[1].document[0].title, "Macro is missing note")
            elif isinstance(doc, VariableDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for variable")
                self.assertEqual("data", element.document[0].title, "Wrong directive type for variable")
            elif isinstance(doc, GenericCommandDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for generic command")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for generic command")
                self.assertEqual("warning", element.document[1].document[0].title, "Generic command is missing warning")
            elif isinstance(doc, ClassDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for class")
                self.assertEqual("py:class", element.document[0].title, "Wrong directive type for class")
            elif isinstance(doc, TestDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for test")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for test")
            elif isinstance(doc, SectionDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for test section")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for test section")
            elif isinstance(doc, CTestDocumentation):
                self.assertIsInstance(element, Directive, "Wrong RST element generated for CTest test")
                self.assertEqual("function", element.document[0].title, "Wrong directive type for CTest test")
            else:
                self.fail(f"Unknown documentation type: {doc}")

    def test_incorrect_variable_type(self):
        var_doc = VariableDocumentation("name", "Not a valid variable type", "0", "This should fail")
        self.assertRaises(ValueError, var_doc.process,
                          RSTWriter("test_writer"))


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains config objects, their defaults,
the config template for confuse, and a function
to convert a dictionary to a :class:`~Settings` object.

The settings object mirrors the template and
the :code:`config_default.yaml` file. Visit that file
for information on what each setting does.
"""

import confuse
import os
from typing import List
from dataclasses import dataclass, field


def config_template(output_dir_relative_to_config: bool = False) -> dict:
    """
    Generates the config template.

    :param output_dir_relative_to_config: Selects whether output.dir should be resolved relative to the config source,
    if not then it is resolved relative to the current working directory.
    """
    return {
        "input": {
            "include_undocumented_function": bool,
            "include_undocumented_macro": bool,
            "include_undocumented_cpp_class": bool,
            "include_undocumented_cpp_attr": bool,
            "include_undocumented_cpp_constructor": bool,
            "include_undocumented_cpp_member": bool,
            "include_undocumented_ct_add_test": bool,
            "include_undocumented_add_test": bool,
            "include_undocumented_ct_add_section": bool,
            "include_undocumented_option": bool,
            "auto_exclude_directories_without_cmake": bool,
            "kwargs_doc_trigger_string": confuse.Optional(confuse.String(), default=":keyword"),
            "exclude_filters": confuse.Optional(list, default=()),
            "function_parameter_name_strip_regex": confuse.Optional(confuse.String(), default=""),
            "macro_parameter_name_strip_regex": confuse.Optional(confuse.String(), default=""),
            "member_parameter_name_strip_regex": confuse.Optional(confuse.String(), default=""),
            "recursive": bool,
            "follow_symlinks": bool
        },
        "output": {
            "directory": confuse.Optional(
                confuse.Filename(cwd=os.getcwd()) if not output_dir_relative_to_config
                else confuse.Filename(in_source_dir=True)),
            "relative_to_config": bool
        },
        "logging": confuse.TypeTemplate(dict),
        "rst": {
            "file_extensions_in_titles": bool,
            "file_extensions_in_modules": bool,
            "module_path_separator": ".",
            "headers": confuse.StrSeq(),
            "prefix": confuse.Optional(confuse.String())
        }

    }


@dataclass
class InputSettings:
    include_undocumented_function: bool = True
    include_undocumented_macro: bool = True
    include_undocumented_cpp_class: bool = True
    include_undocumented_cpp_attr: bool = True
    include_undocumented_cpp_constructor: bool = True
    include_undocumented_cpp_member: bool = True
    include_undocumented_ct_add_test: bool = True
    include_undocumented_ct_add_section: bool = True
    include_undocumented_add_test: bool = True
    include_undocumented_option: bool = True
    auto_exclude_directories_without_cmake: bool = True
    kwargs_doc_trigger_string: str = ":param **kwargs:"
    exclude_filters: List[str] = ()
    function_parameter_name_strip_regex: str = ""
    macro_parameter_name_strip_regex: str = ""
    member_parameter_name_strip_regex: str = ""
    recursive: bool = False
    follow_symlinks: bool = False


@dataclass
class OutputSettings:
    directory: str = None
    relative_to_config: bool = False


@dataclass
class LoggingSettings:
    logger_config: dict = field(default_factory=lambda: ({}))


@dataclass
class RSTSettings:
    file_extensions_in_titles: bool = False
    file_extensions_in_modules: bool = False
    prefix: str = None
    module_path_separator: str = "."
    headers: List[str] = ('#', '*', '=', '-', '_', '~', '!', '&', '@', '^')


@dataclass
class Settings:
    input: InputSettings = field(default_factory=lambda: InputSettings())
    output: OutputSettings = field(default_factory=lambda: OutputSettings())
    logging: LoggingSettings = field(default_factory=lambda: LoggingSettings())
    rst: RSTSettings = field(default_factory=lambda: RSTSettings())


def dict_to_settings(input_dict: dict) -> Settings:
    input_settings = InputSettings(**input_dict["input"])
    output_settings = OutputSettings(**input_dict["output"])
    logging_settings = LoggingSettings(input_dict["logging"])
    rst_settings = RSTSettings(**input_dict["rst"])
    return Settings(
        input_settings,
        output_settings,
        logging_settings,
        rst_settings)



# -*- coding: utf-8 -*-
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import cminx
import datetime
import os
import subprocess
import sys

######################################
#  Work out paths this script needs  #
######################################

now = datetime.datetime.now()
src_dir  = os.path.abspath(os.path.dirname(__file__))
docs_dir = os.path.abspath(os.path.join(src_dir, ".."))
root_dir = os.path.join(docs_dir, "..")

############################################
#  Gather some metadata about the project  #
############################################

project = 'CMinx'
author = 'CMakePP Team'

# Get the version from the git tag
git_cmd = ['git', 'describe', '--tags', '--abbrev=0']

# N.B. release includes alpha, beta, rc, etc.; version is pure numeric
release = subprocess.check_output(git_cmd).strip().decode()
version = release[1:]

#########################
#  Run CMinx on itself  #
#########################

cminx_out_dir = os.path.join(src_dir, "developer", "cmake")
cminx_in_dir = os.path.join(root_dir, "cmake")
args = ["-r", "-o", cminx_out_dir, cminx_in_dir]
cminx.main(args)

################################################################################
#             Shouldn't need to modify anything below this point               #
################################################################################

copyright = '{}, {}'.format(now.year, author)

# -- General configuration -----------------------------------------------------
templates_path = ['.templates']
source_suffix = '.rst'
master_doc = 'index'
numfig = True
exclude_patterns = ['build', '.templates']
pygments_style = 'sphinx'
# Required theme setup
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'vcs_pageview_mode' : 'edit'
}
html_context = {
    'display_github' : True,
    'github_user' : 'CMakePP',
    'github_repo' : 'CMinx',
    'github_version' : 'master/docs/source/'
}
htmlhelp_basename = project + 'doc'
extensions = [
    'sphinx.ext.mathjax',
    'sphinx.ext.githubpages',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx_rtd_theme'

]
autosummary_generate = True

epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
import sys
import unittest

import context
import helpers

import cminx
from cminx.config import Settings, InputSettings, OutputSettings, RSTSettings, LoggingSettings


class TestInit(unittest.TestCase):
    """
    Tests the functions found in __init__.py
    """

    def setUp(self):
        self.cwd = os.path.abspath(os.path.dirname(__file__))
        self.input_dir = context.example_dir
        self.output_dir = os.path.join(self.cwd, "output")
        self.input_file = context.example_cmake
        self.output_file = os.path.join(self.output_dir, "example.rst")
        self.output_index_file = os.path.join(self.output_dir, "index.rst")
        self.maxDiff = None

    def tearDown(self):
        try:
            shutil.rmtree(self.output_dir)
            pass
        except FileNotFoundError:
            pass  # Test just didn't write to the directory

    def test_document(self):
        """Tests the document command in the CMinx module"""
        sys.stdout = open(os.devnull, 'w')
        cminx.document(self.input_file,
                       Settings(InputSettings(recursive=True), OutputSettings(directory=self.output_dir)))
        sys.stdout = sys.__stdout__

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that the file is in the directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

    def test_document_prefix(self):
        """Tests the document command in the CMinx module with the prefix option"""
        sys.stdout = open(os.devnull, 'w')
        cminx.document(self.input_file,
                       Settings(input=InputSettings(recursive=True), output=OutputSettings(directory=self.output_dir),
                                rst=RSTSettings(prefix=context.prefix)))
        sys.stdout = sys.__stdout__

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that the file is in the directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

    def test_header_extensions_no_undocumented_diff_header(self):
        input_settings = InputSettings(include_undocumented_ct_add_test=False, include_undocumented_cpp_attr=False,
                                       include_undocumented_function=False, include_undocumented_cpp_class=False,
                                       include_undocumented_macro=False, include_undocumented_cpp_member=False,
                                       include_undocumented_ct_add_section=False, include_undocumented_add_test=False,
                                       recursive=True)
        output_settings = OutputSettings(directory=self.output_dir)
        rst_settings = RSTSettings(file_extensions_in_modules=True, file_extensions_in_titles=True,
                                   headers=['^', '*', '=', '-', '_', '~', '!', '&', '@'])
        settings = Settings(input_settings, output_settings, LoggingSettings(), rst_settings)
        cminx.document(self.input_dir, settings)
        # diff = helpers.diff_files(self.output_file, context.corr_example_no_undocumented_diff_header)
        # self.assertEqual(diff, "")
        with open(context.corr_example_no_undocumented_diff_header, 'r') as corr_file, \
                open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_recursive(self):
        """Tests the use of CMinx in recursive mode"""
        args = ["-r", "-o", self.output_dir, self.input_dir]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that reST file is in top-level directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

        with open(context.corr_example_rst, 'r') as corr_file, open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

        with open(context.corr_index_rst, 'r') as corr_file, open(self.output_index_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_recursive_prefix(self):
        """Tests the use of CMinx in recursive mode with a prefix"""
        args = ["-r", "-p", context.prefix, "-o", self.output_dir, self.input_dir]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that reST file is in top-level directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

        with open(context.corr_example_prefix_rst, 'r') as corr_file, open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

        with open(context.corr_index_prefix_rst, 'r') as corr_file, open(self.output_index_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_nonexistent_input(self):
        """Tests whether trying to document a nonexistent file exits with -1"""
        args = ["nonexistent-file.cmake"]
        exit_code = None
        try:
            helpers.quiet_cminx(args)
        except SystemExit as e:
            exit_code = e.code

        self.assertEqual(-1, exit_code, f"Exit code for nonexistent input incorrect (None means did not exit):"
                                        "{exit_code}")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module and associated classes provide a pure-Python interface for generating reStructuredText.
The generated text may be either consumed by another Python application or directly written to a file-like object.
This module does not currently perform syntax validation, so it is up to the application developer to
not generate invalid RST structures.


:Author: Branden Butler
:License: Apache 2.0
"""
from enum import Enum
from typing import Any, Union, IO, List, Tuple

from cminx import Settings


class ListType(Enum):
    """
    Represents the different types of RST
    lists.
    """

    ENUMERATED = 0
    """
    An enumerated RST list.
    """
    BULLETED = 1
    """
    A bulleted RST list. 
    """


def interpreted_text(role: str, text: str) -> str:
    """
    This function generates an interpreted text RST element
    from the specified text and using the specified role.
    Interpreted text may appear almost anywhere regular text can
    so making it a method of :class:`RSTWriter` would limit its functionality.


    :param role: The role of the interpreted text
    :param text: The literal text of the interpreted text
    :return: The interpreted text element as a string
    """
    return f":{role}:`{text}`"


class Paragraph(object):
    """
    Represents an RST paragraph
    """

    def __init__(self, text: str, indent: str = "") -> None:
        self.text: str = text
        self.prefix: str = indent
        self.text_string: str = ""
        self.build_text_string()

    def build_text_string(self) -> None:
        """
        Populates Paragraph.text_string with the
        RST string corresponding to this paragraph
        """
        self.text_string = "\n".join(
            [self.prefix + text for text in self.text.split("\n")])

    def __str__(self) -> str:
        return self.text_string


class Field(object):
    """
    Represents an RST field, such as Author
    """

    def __init__(self, field_name: str, field_text: str, indent: str = "") -> None:
        self.field_name: str = field_name
        self.field_text: str = field_text
        self.field_string: str = ""
        self.indent: str = indent
        self.build_field_string()

    def build_field_string(self) -> None:
        """
        Populates Field.field_string with the
        RST string corresponding to this field
        """

        self.field_string = f"\n{self.indent}:{self.field_name}: {self.field_text}"

    def __str__(self) -> str:
        return self.field_string


class DocTest(object):
    """
    Represents an RST DocTest
    """

    def __init__(self, test_line: str, expected_output: str, indent: str = "") -> None:
        self.test_line: str = test_line
        self.expected_output: str = expected_output
        self.doctest_string: str = ""
        self.indent: str = indent
        self.build_doctest_string()

    def build_doctest_string(self) -> None:
        """
        Populates DocTest.doctest_string with the
        RST string corresponding to this DocTest
        """
        self.doctest_string = f"\n{self.indent}>>> {self.test_line}\n{self.expected_output}\n"

    def __str__(self) -> str:
        return self.doctest_string


class RSTList(object):
    """
    Represents one of the two types of RST lists:
    Enumerated or Bulleted
    """

    def __init__(self, items: Tuple[str], list_type: ListType, indent: str = "") -> None:
        self.items: Tuple[str] = items
        self.list_type: ListType = list_type
        self.list_string: str = ""
        self.indent: str = indent
        self.build_list_string()

    def build_list_string(self) -> None:
        """
        Populates RSTList.list_string with the
        RST string corresponding to this list.

        :raises ValueError: When :py:attr:`~.RSTList.list_type` is unknown.
        """

        self.list_string = "\n"
        if self.list_type == ListType.ENUMERATED:
            for i in range(0, len(self.items)):
                self.list_string += f"{self.indent}{i + 1}. {self.items[i]}\n"
        elif self.list_type == ListType.BULLETED:
            for item in self.items:
                self.list_string += f"{self.indent}* {item}\n"
        else:
            raise ValueError("Unknown list type")

    def __str__(self) -> str:
        return self.list_string


class Heading(object):
    """
    Represents a section heading
    """

    def __init__(self, title: str, header_char: str) -> None:
        self.title = title
        self.header_char = header_char
        self.heading_string = ""
        self.build_heading_string()

    def build_heading_string(self) -> None:
        """
        Populates Heading.heading_string with the
        RST string corresponding to this heading
        """
        heading = ""
        for _ in self.title:
            heading += self.header_char
        self.heading_string = f"\n{heading}\n{self.title}\n{heading}"

    def __str__(self) -> str:
        return self.heading_string


class SimpleTable(object):
    """
    Represents an RST simple table.
    """

    def __init__(self, tab: List[List[str]], headings: List[str]):
        self.table: List[List[str]] = tab
        self.column_headings: List[str] = headings
        self.table_string: str = ""
        self.build_table_string()

    def build_table_string(self) -> None:
        """
        Populates SimpleTable.table_string with the
        RST string equivalent of this table
        """
        # Easiest way to do this is loop over all cells and find the longest element, use that to compute the row
        # separator width, and then loop over cells again to add them in with proper separation. Fastest way is to
        # compute the separator width, then use list comprehension to build the table string. Probably going to be
        # much less readable. Performance differences will be minimal unless we're building tables with thousands of
        # cells.
        row_separator = ""
        row_separator_width = 0

        # Used to ensure number of columns is consistent throughout the table;
        # Length of row gives number of cells in row, which corresponds to
        # number of columns
        num_columns = len(self.table[0])

        if len(self.column_headings) != 0 and len(
                self.column_headings) != num_columns:
            raise ValueError(
                "Different number of headings than number of columns in table")

        # Find the longest cell to use as the row separator width. Yes I know doing it this way is inefficient,
        # but readability is more important unless we're dealing with thousands
        # of cells.
        for row in self.table:
            # Check column number against number of cells in row; should be
            # equal
            if len(row) != num_columns:
                raise ValueError("Number of columns in table is inconsistent")
            for cell in row:
                # Temporarily cache in case calculating length is
                # time-consuming.
                cell_length = len(str(cell))
                row_separator_width = cell_length if cell_length > row_separator_width else row_separator_width

        # Account for column headings, in case they're longer than the cells
        for heading in self.column_headings:
            length = len(heading)  # Cache so only need to compute length once
            row_separator_width = length if length > row_separator_width else row_separator_width

        # We know the width, now to construct the separator itself
        for i in range(row_separator_width):
            row_separator += "="

        table_str = ""
        heading_lines = {"overline": "", "headings": "", "underline": ""}
        # Build heading overlines/underlines and add correct spacing
        for heading in self.column_headings:
            heading_len = len(heading)
            for i in range(0, row_separator_width):
                heading_lines["overline"] += '='

                # If current line position is not greater than the length of the heading, add the character at position
                # Else, add spaces until end of row separator
                if i < heading_len:
                    heading_lines["headings"] += heading[i]
                else:
                    heading_lines["headings"] += " "
            heading_lines["overline"] += "  "  # Recommended 2 spaces between columns
            heading_lines["headings"] += "  "

        # Overline and underline should be the same
        heading_lines["underline"] = heading_lines["overline"]
        table_str += "\n".join(heading_lines.values()) + "\n"

        # Add cells to table string
        for row in self.table:
            row_str = ""
            for cell in row:
                row_str += str(cell)
                for i in range(0, row_separator_width - len(str(cell))):
                    row_str += " "
                row_str += "  "
            table_str += row_str + "\n"
        for _ in self.table[0]:
            table_str += row_separator + "  "
        table_str += "\n"
        self.table_string = table_str

    def __str__(self) -> str:
        return self.table_string


class DirectiveHeading(object):
    """
    Represents the unique heading for a Directive (.. :<name>:)
    """

    def __init__(self, title: str, indent: str, args: str) -> None:
        self.title: str = title
        self.indent: str = indent
        self.args: str = args
        self.heading_string: str = ""
        self.build_heading_string()

    def build_heading_string(self) -> None:
        """Builds the directive heading string and stores in :py:attr:`~heading_str`"""
        self.heading_string = f"\n{self.indent}.. {self.title}:: {self.args}"

    def __str__(self) -> str:
        return self.heading_string


class Option(object):
    """
    Represents a directive option, such as maxdepth
    """

    def __init__(self, name: str, value: str, indent: str) -> None:
        self.name: str = name
        self.value: str = value
        self.indent: str = indent
        self.option_string: str = ""
        self.build_option_string()

    def build_option_string(self) -> None:
        """Builds the option string and stores in :py:attr:`~option_string`"""
        self.option_string = f"{self.indent}:{self.name}: {self.value}"

    def __str__(self) -> str:
        return self.option_string


def get_indents(num: int) -> str:
    """
    Get the string containing the necessary indent.
    The number of spaces in the returned string equals three times
    the input number, because directives require three spaces
    so the text lines up with the first letter of the directive name.

    :return: A string containing the correct number of space characters, derived from the indent level.
    """
    indents = ""
    for i in range(0, num):
        # Directives require the first non-whitespace character
        # of every line to line up with the first letter of
        # the directive name
        indents += '   '
    return indents


class RSTWriter(object):
    """
    Base reStructuredText writer. Does not perform verification.
    This class presents an object-oriented and procedural API
    for generating reStructuredText documents. The document is
    held in an intermediate object-oriented representation until
    the text representation is requested either through
    :py:meth:`~cminx.rstwriter.RSTWriter.to_text`
    or by calling :py:func:`str` on this object.
    """

    heading_level_chars: List[str] = ['#', '*', '=', '-', '_', '~', '!', '&', '@', '^']
    """Characters to use as heading overline/underline, indexed by section_level"""

    def __init__(
            self,
            title: str,
            section_level: int = 0,
            settings: Settings = Settings(), indent: int = 0) -> None:
        self.__title: str = title
        self.section_level: int = section_level
        self.settings: Settings = settings
        if settings.rst.headers is not None:
            self.heading_level_chars = settings.rst.headers

        self.indent: int = indent
        self.header_char: str = self.heading_level_chars[section_level]
        # Heading must be first in the document tree
        self.document: List[Any] = [self.build_heading()]

    def clear(self) -> None:
        """
        Clear all document tree elements (besides required heading)
        """
        del self.document[1:]

    @property
    def title(self) -> str:
        return self.__title

    @title.setter
    def title(self, new_title: str) -> None:
        """
        Rebuild heading and replace the old one in the document tree whenever title is changed

        :param new_title: The new section title to be used.
        """
        self.__title = new_title
        self.document[0] = self.build_heading()

    def bulleted_list(self, *items: str) -> None:
        """
        Add a bulleted list to the document tree.

        :param items: varargs containing the desired list items.
        """

        self.document.append(RSTList(items, ListType.BULLETED, indent=get_indents(self.indent)))

    def enumerated_list(self, *items: str) -> None:
        """
        Add an enumerated list to the document tree; e.g.
            1. Item 1
            2. Item 2

        :param items: varargs containing the desired list items.
        """
        self.document.append(RSTList(items, ListType.ENUMERATED, indent=get_indents(self.indent)))

    def field(self, field_name: str, field_text: str) -> None:
        """
        Add a field, such as Author or Version, to the document tree.

        :param field_name: Name of the field being added, such as Author.

        :param field_text: Value of the field, such as the author's name.
        """
        self.document.append(Field(field_name, field_text, get_indents(self.indent)))

    def doctest(self, test_line: str, expected_output: str) -> None:
        """
        Adds a doctest segment to the document tree.
        A doctest segment appears as an interactive python session.
        The doctest Python module can then be used to scan for these segments
        and execute the test_line to ensure the output is the same.

        :param test_line: Python code segment being used as the line to be tested.

        :param expected_output: The exact string that is expected to be returned when test_line is evaluated.
        """
        self.document.append(DocTest(test_line, expected_output, indent=get_indents(self.indent)))

    def section(self, title: str) -> 'RSTWriter':
        """
        Constructs another RSTWriter and adds it to the document before returning it for use

        :param title: The heading title to be used for the new subsection.
        """
        sect = RSTWriter(
            title,
            section_level=self.section_level + 1,
            settings=self.settings)
        self.document.append(sect)
        return sect

    def text(self, txt: str) -> None:
        """
        Add a paragraph to document tree.

        :param txt: The content of the new paragraph.
        """
        self.document.append(Paragraph(txt, indent=get_indents(self.indent)))

    def directive(self, name: str, *arguments: str) -> 'Directive':
        """
        Construct a directive and return it

        :param name: Name of the directive being used, such as toctree or admonition.

        :param arguments: Varargs to be used as the directive arguments, such as a topic title.
        """
        w = Directive(name, self.indent, *arguments, settings=self.settings)
        self.document.append(w)
        return w

    def build_heading(self) -> Heading:
        """
        Adds overline and underline to RSTWriter.title (character is RSTWriter.header_char) and returns it

        :return: A fully constructed Heading object.
        """
        return Heading(self.__title, self.header_char)

    def simple_table(self, tab: List[List[str]], column_headings: List[str] = ()):
        """
        Add a simple table to the document tree. tab is a 2-dimensional list containing the table data. The first
        index is the row, the second is column. The first column may not contain multiple lines, all other columns
        may. column_headings is a list where each element is treated as a heading for that specific column,
        i.e. index 0 will be the heading for the leftmost column.

        :param tab: A two-dimensional list representing table data. First index is row, second is column. Each row
        must have the same number of columns.

        :param column_headings: A single-dimensional list containing column headings, if required. Index zero is the
        heading for the left-most column. List must be same length as number of columns.

        :raise ValueError: If tab has inconsistent numbers of columns or column_headings length (if nonzero) does not
        match number of columns in tab.
        """

        self.document.append(SimpleTable(tab, column_headings))

    def to_text(self) -> str:
        """
        Return text representation of this document

        :return: The completed RST document in string form.
        """
        document_string = ""
        for element in self.document:
            document_string += f"{element}\n"
        return document_string

    def __str__(self) -> str:
        """
        Equivalent to :func: `~cminx.RSTWriter.to_text`

        :return: The completed RST document in string form.
        """
        return self.to_text()

    def write_to_file(self, file: Union[str, IO]) -> None:
        """
        Write text representation of this document to file. File may be a string (path will be opened in write mode)
        or a file-like object.

        :param file: File to write to. May be a string representing a real filepath on the local filesystem (will be
        overwritten), or a file-like object.

        :raise ValueError: If file is nothing or empty string.

        :raise TypeError: If file is not str or object with 'write()' method
        """

        # Don't use 'is' to compare to empty string, file might be subclass of str or could be a C-built string,
        # both of which are different references Don't use "==" to compare to None, None will be a singleton in every
        # case and objects may override __eq__()
        if file == "" or file is None:
            raise ValueError("File not set, cannot write RST document")
        if isinstance(file, str):
            # Strip leading/trailing whitespace, so we don't end up with '. '
            # as a file, I have seen it happen before
            with open(file.strip(),
                      'w') as f:
                f.write(str(self))
        else:
            # Might be invalid object, checking to make sure it's file-like
            if hasattr(file, "write"):
                file.write(str(self))
            else:
                # Not file-like object or string
                raise TypeError(
                    f"File is neither a string nor a file-like object, type is {type(file)}, cannot write RST document")


class Directive(RSTWriter):
    """
    Use :func:`cminx.RSTWriter.directive` to construct.
    Represents an RST directive, such as toctree or an admonition.
    Does not verify that the directive or its arguments are valid.
    """

    def __init__(self, name: str, indent: int = 0, *arguments: str, settings: Settings = Settings()) -> None:
        """
        :param name: Name of the directive being constructed, eg. toctree or admonition.

        :param indent: Indent level of this directive. 0 is root, 1 is under another directive, etc.

        :param arguments: Varargs to be used as the directive arguments, eg. a topic's title.
        """

        self.arguments: Tuple[str] = arguments
        self.options: List[Option] = []
        super().__init__(name, settings=settings, indent=indent + 1)

    def build_heading(self) -> DirectiveHeading:
        """
        Build directive heading format (ex. '.. toctree::') and return.

        :return: Correctly formatted directive heading string.
        """
        return DirectiveHeading(
            self.title, get_indents(
                self.indent - 1), self.format_arguments())

    def format_arguments(self) -> str:
        """
        Format argument list into the correct argument string for use with directives.

        :return: A string representing the directive arguments.
        """
        return ','.join(map(str, self.arguments))

    def option(self, name: str, value: str = "") -> None:
        """
        Add an option, such as toctree's maxdepth. Does not verify if valid option
        """
        self.options.append(
            Option(
                name,
                value,
                get_indents(
                    self.indent)))

    def to_text(self) -> str:
        """
        Return text representation of this document

        :return: The completed RST document in string form.
        """
        document_string = f"{self.document[0]}\n"

        for option in self.options:
            document_string += f"{option}\n"

        if len(self.document) > 1:
            document_string += "\n"

        for element in self.document[1:]:
            document_string += f"{element}\n"
        return document_string



# Generated from CMake.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CMakeParser import CMakeParser
else:
    from CMakeParser import CMakeParser

# This class defines a complete generic visitor for a parse tree produced by CMakeParser.

class CMakeVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CMakeParser#cmake_file.
    def visitCmake_file(self, ctx:CMakeParser.Cmake_fileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#documented_command.
    def visitDocumented_command(self, ctx:CMakeParser.Documented_commandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#documented_module.
    def visitDocumented_module(self, ctx:CMakeParser.Documented_moduleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#bracket_doccomment.
    def visitBracket_doccomment(self, ctx:CMakeParser.Bracket_doccommentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#command_invocation.
    def visitCommand_invocation(self, ctx:CMakeParser.Command_invocationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#single_argument.
    def visitSingle_argument(self, ctx:CMakeParser.Single_argumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CMakeParser#compound_argument.
    def visitCompound_argument(self, ctx:CMakeParser.Compound_argumentContext):
        return self.visitChildren(ctx)



del CMakeParser


# Generated from CMake.g4 by ANTLR 4.7.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\21")
        buf.write("<\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\3\2\5\2\22\n\2\3\2\3\2\3\2\7\2\27\n\2\f\2\16\2\32")
        buf.write("\13\2\3\2\3\2\3\3\3\3\3\3\3\4\3\4\3\5\3\5\3\6\3\6\3\6")
        buf.write("\3\6\7\6)\n\6\f\6\16\6,\13\6\3\6\3\6\3\7\3\7\3\b\3\b\3")
        buf.write("\b\7\b\65\n\b\f\b\16\b8\13\b\3\b\3\b\3\b\2\2\t\2\4\6\b")
        buf.write("\n\f\16\2\3\4\2\t\n\f\r\2<\2\21\3\2\2\2\4\35\3\2\2\2\6")
        buf.write(" \3\2\2\2\b\"\3\2\2\2\n$\3\2\2\2\f/\3\2\2\2\16\61\3\2")
        buf.write("\2\2\20\22\5\6\4\2\21\20\3\2\2\2\21\22\3\2\2\2\22\30\3")
        buf.write("\2\2\2\23\27\5\4\3\2\24\27\5\n\6\2\25\27\5\b\5\2\26\23")
        buf.write("\3\2\2\2\26\24\3\2\2\2\26\25\3\2\2\2\27\32\3\2\2\2\30")
        buf.write("\26\3\2\2\2\30\31\3\2\2\2\31\33\3\2\2\2\32\30\3\2\2\2")
        buf.write("\33\34\7\2\2\3\34\3\3\2\2\2\35\36\5\b\5\2\36\37\5\n\6")
        buf.write("\2\37\5\3\2\2\2 !\7\5\2\2!\7\3\2\2\2\"#\7\6\2\2#\t\3\2")
        buf.write("\2\2$%\7\t\2\2%*\7\3\2\2&)\5\f\7\2\')\5\16\b\2(&\3\2\2")
        buf.write("\2(\'\3\2\2\2),\3\2\2\2*(\3\2\2\2*+\3\2\2\2+-\3\2\2\2")
        buf.write(",*\3\2\2\2-.\7\4\2\2.\13\3\2\2\2/\60\t\2\2\2\60\r\3\2")
        buf.write("\2\2\61\66\7\3\2\2\62\65\5\f\7\2\63\65\5\16\b\2\64\62")
        buf.write("\3\2\2\2\64\63\3\2\2\2\658\3\2\2\2\66\64\3\2\2\2\66\67")
        buf.write("\3\2\2\2\679\3\2\2\28\66\3\2\2\29:\7\4\2\2:\17\3\2\2\2")
        buf.write("\t\21\26\30(*\64\66")
        return buf.getvalue()


class CMakeParser ( Parser ):

    grammarFileName = "CMake.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'('", "')'", "<INVALID>", "<INVALID>", 
                     "'#[[['", "'#]]'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "Module_docstring", 
                      "Docstring", "Doccomment_start", "Blockcomment_end", 
                      "Identifier", "Unquoted_argument", "Escape_sequence", 
                      "Quoted_argument", "Bracket_argument", "Bracket_comment", 
                      "Line_comment", "Newline", "Space" ]

    RULE_cmake_file = 0
    RULE_documented_command = 1
    RULE_documented_module = 2
    RULE_bracket_doccomment = 3
    RULE_command_invocation = 4
    RULE_single_argument = 5
    RULE_compound_argument = 6

    ruleNames =  [ "cmake_file", "documented_command", "documented_module", 
                   "bracket_doccomment", "command_invocation", "single_argument", 
                   "compound_argument" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    Module_docstring=3
    Docstring=4
    Doccomment_start=5
    Blockcomment_end=6
    Identifier=7
    Unquoted_argument=8
    Escape_sequence=9
    Quoted_argument=10
    Bracket_argument=11
    Bracket_comment=12
    Line_comment=13
    Newline=14
    Space=15

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class Cmake_fileContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(CMakeParser.EOF, 0)

        def documented_module(self):
            return self.getTypedRuleContext(CMakeParser.Documented_moduleContext,0)


        def documented_command(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Documented_commandContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Documented_commandContext,i)


        def command_invocation(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Command_invocationContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Command_invocationContext,i)


        def bracket_doccomment(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Bracket_doccommentContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Bracket_doccommentContext,i)


        def getRuleIndex(self):
            return CMakeParser.RULE_cmake_file

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCmake_file" ):
                listener.enterCmake_file(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCmake_file" ):
                listener.exitCmake_file(self)




    def cmake_file(self):

        localctx = CMakeParser.Cmake_fileContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_cmake_file)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 15
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==CMakeParser.Module_docstring:
                self.state = 14
                self.documented_module()


            self.state = 22
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==CMakeParser.Docstring or _la==CMakeParser.Identifier:
                self.state = 20
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
                if la_ == 1:
                    self.state = 17
                    self.documented_command()
                    pass

                elif la_ == 2:
                    self.state = 18
                    self.command_invocation()
                    pass

                elif la_ == 3:
                    self.state = 19
                    self.bracket_doccomment()
                    pass


                self.state = 24
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 25
            self.match(CMakeParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Documented_commandContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def bracket_doccomment(self):
            return self.getTypedRuleContext(CMakeParser.Bracket_doccommentContext,0)


        def command_invocation(self):
            return self.getTypedRuleContext(CMakeParser.Command_invocationContext,0)


        def getRuleIndex(self):
            return CMakeParser.RULE_documented_command

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDocumented_command" ):
                listener.enterDocumented_command(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDocumented_command" ):
                listener.exitDocumented_command(self)




    def documented_command(self):

        localctx = CMakeParser.Documented_commandContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_documented_command)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 27
            self.bracket_doccomment()
            self.state = 28
            self.command_invocation()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Documented_moduleContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Module_docstring(self):
            return self.getToken(CMakeParser.Module_docstring, 0)

        def getRuleIndex(self):
            return CMakeParser.RULE_documented_module

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDocumented_module" ):
                listener.enterDocumented_module(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDocumented_module" ):
                listener.exitDocumented_module(self)




    def documented_module(self):

        localctx = CMakeParser.Documented_moduleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_documented_module)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self.match(CMakeParser.Module_docstring)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Bracket_doccommentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Docstring(self):
            return self.getToken(CMakeParser.Docstring, 0)

        def getRuleIndex(self):
            return CMakeParser.RULE_bracket_doccomment

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBracket_doccomment" ):
                listener.enterBracket_doccomment(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBracket_doccomment" ):
                listener.exitBracket_doccomment(self)




    def bracket_doccomment(self):

        localctx = CMakeParser.Bracket_doccommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_bracket_doccomment)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 32
            self.match(CMakeParser.Docstring)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Command_invocationContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Identifier(self):
            return self.getToken(CMakeParser.Identifier, 0)

        def single_argument(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Single_argumentContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Single_argumentContext,i)


        def compound_argument(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Compound_argumentContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Compound_argumentContext,i)


        def getRuleIndex(self):
            return CMakeParser.RULE_command_invocation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCommand_invocation" ):
                listener.enterCommand_invocation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCommand_invocation" ):
                listener.exitCommand_invocation(self)




    def command_invocation(self):

        localctx = CMakeParser.Command_invocationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_command_invocation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 34
            self.match(CMakeParser.Identifier)
            self.state = 35
            self.match(CMakeParser.T__0)
            self.state = 40
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << CMakeParser.T__0) | (1 << CMakeParser.Identifier) | (1 << CMakeParser.Unquoted_argument) | (1 << CMakeParser.Quoted_argument) | (1 << CMakeParser.Bracket_argument))) != 0):
                self.state = 38
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [CMakeParser.Identifier, CMakeParser.Unquoted_argument, CMakeParser.Quoted_argument, CMakeParser.Bracket_argument]:
                    self.state = 36
                    self.single_argument()
                    pass
                elif token in [CMakeParser.T__0]:
                    self.state = 37
                    self.compound_argument()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 42
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 43
            self.match(CMakeParser.T__1)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Single_argumentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Identifier(self):
            return self.getToken(CMakeParser.Identifier, 0)

        def Unquoted_argument(self):
            return self.getToken(CMakeParser.Unquoted_argument, 0)

        def Bracket_argument(self):
            return self.getToken(CMakeParser.Bracket_argument, 0)

        def Quoted_argument(self):
            return self.getToken(CMakeParser.Quoted_argument, 0)

        def getRuleIndex(self):
            return CMakeParser.RULE_single_argument

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSingle_argument" ):
                listener.enterSingle_argument(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSingle_argument" ):
                listener.exitSingle_argument(self)




    def single_argument(self):

        localctx = CMakeParser.Single_argumentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_single_argument)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 45
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << CMakeParser.Identifier) | (1 << CMakeParser.Unquoted_argument) | (1 << CMakeParser.Quoted_argument) | (1 << CMakeParser.Bracket_argument))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Compound_argumentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def single_argument(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Single_argumentContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Single_argumentContext,i)


        def compound_argument(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CMakeParser.Compound_argumentContext)
            else:
                return self.getTypedRuleContext(CMakeParser.Compound_argumentContext,i)


        def getRuleIndex(self):
            return CMakeParser.RULE_compound_argument

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCompound_argument" ):
                listener.enterCompound_argument(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCompound_argument" ):
                listener.exitCompound_argument(self)




    def compound_argument(self):

        localctx = CMakeParser.Compound_argumentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_compound_argument)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 47
            self.match(CMakeParser.T__0)
            self.state = 52
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << CMakeParser.T__0) | (1 << CMakeParser.Identifier) | (1 << CMakeParser.Unquoted_argument) | (1 << CMakeParser.Quoted_argument) | (1 << CMakeParser.Bracket_argument))) != 0):
                self.state = 50
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [CMakeParser.Identifier, CMakeParser.Unquoted_argument, CMakeParser.Quoted_argument, CMakeParser.Bracket_argument]:
                    self.state = 48
                    self.single_argument()
                    pass
                elif token in [CMakeParser.T__0]:
                    self.state = 49
                    self.compound_argument()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 54
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 55
            self.match(CMakeParser.T__1)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
import unittest

import context
import helpers


class TestConfig(unittest.TestCase):
    """
    Tests the functionality of the configuration system
    """

    def setUp(self):
        self.cwd = os.path.abspath(os.path.dirname(__file__))
        self.input_dir = context.example_dir
        self.output_dir = os.path.join(self.cwd, "output")
        self.input_file = context.example_cmake
        self.output_file = os.path.join(self.output_dir, "example.rst")
        self.output_index_file = os.path.join(self.output_dir, "index.rst")
        self.config_dir = context.config_dir
        self.maxDiff = None

    def tearDown(self):
        try:
            shutil.rmtree(self.output_dir)
            pass
        except FileNotFoundError:
            pass  # Test just didn't write to the directory

    def test_no_undocumented(self):
        """Tests include_undocumented_* settings as False"""
        args = ["-r", "-o", self.output_dir, "-s", context.no_include_undocumented_config, self.input_dir]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that reST file is in top-level directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

        with open(context.corr_example_no_undocumented, 'r') as corr_file, open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

        with open(context.corr_index_rst, 'r') as corr_file, open(self.output_index_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_no_auto_exclude(self):
        """Tests auto_exclude_directories_as_cmake setting as False"""
        args = ["-r", "-o", self.output_dir, "-s", context.no_auto_exclude_config, self.input_dir]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that reST file is in top-level directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

        with open(context.corr_example_rst, 'r') as corr_file, open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

        with open(context.corr_index_no_auto_exclude, 'r') as corr_file, open(self.output_index_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_exclude_filters(self):
        """Tests the gitignore-like exclusion filters"""
        args = ["-r", "-o", self.output_dir, "-s", context.exclusion_filters_config, self.input_dir]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was found
        is_dir = os.path.isdir(self.output_dir)
        self.assertTrue(is_dir, "Output directory structure incorrect")

        # Test that reST file is in top-level directory
        is_file = os.path.isfile(self.output_file)
        self.assertTrue(is_file, "Output file does not exist")

        with open(context.corr_example_rst, 'r') as corr_file, open(self.output_file, 'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

        with open(context.corr_index_exclusion_filters, 'r') as corr_file, open(self.output_index_file,
                                                                              'r') as generated_file:
            self.assertEqual(corr_file.read(), generated_file.read())

    def test_exclude_input(self):
        """Tests the gitignore-like exclusion filters"""
        args = ["-o", self.output_dir, "-s", context.exclude_input_file_config, self.input_file]
        helpers.quiet_cminx(args)

        # Test that the top-level directory was not found, since we don't do anything
        is_dir = os.path.exists(self.output_dir)
        self.assertFalse(is_dir, "Output directory structure created when input should've been ignored")



if __name__ == '__main__':
    unittest.main()



# Generated from CMake.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CMakeParser import CMakeParser
else:
    from CMakeParser import CMakeParser

# This class defines a complete listener for a parse tree produced by CMakeParser.
class CMakeListener(ParseTreeListener):

    # Enter a parse tree produced by CMakeParser#cmake_file.
    def enterCmake_file(self, ctx:CMakeParser.Cmake_fileContext):
        pass

    # Exit a parse tree produced by CMakeParser#cmake_file.
    def exitCmake_file(self, ctx:CMakeParser.Cmake_fileContext):
        pass


    # Enter a parse tree produced by CMakeParser#documented_command.
    def enterDocumented_command(self, ctx:CMakeParser.Documented_commandContext):
        pass

    # Exit a parse tree produced by CMakeParser#documented_command.
    def exitDocumented_command(self, ctx:CMakeParser.Documented_commandContext):
        pass


    # Enter a parse tree produced by CMakeParser#documented_module.
    def enterDocumented_module(self, ctx:CMakeParser.Documented_moduleContext):
        pass

    # Exit a parse tree produced by CMakeParser#documented_module.
    def exitDocumented_module(self, ctx:CMakeParser.Documented_moduleContext):
        pass


    # Enter a parse tree produced by CMakeParser#bracket_doccomment.
    def enterBracket_doccomment(self, ctx:CMakeParser.Bracket_doccommentContext):
        pass

    # Exit a parse tree produced by CMakeParser#bracket_doccomment.
    def exitBracket_doccomment(self, ctx:CMakeParser.Bracket_doccommentContext):
        pass


    # Enter a parse tree produced by CMakeParser#command_invocation.
    def enterCommand_invocation(self, ctx:CMakeParser.Command_invocationContext):
        pass

    # Exit a parse tree produced by CMakeParser#command_invocation.
    def exitCommand_invocation(self, ctx:CMakeParser.Command_invocationContext):
        pass


    # Enter a parse tree produced by CMakeParser#single_argument.
    def enterSingle_argument(self, ctx:CMakeParser.Single_argumentContext):
        pass

    # Exit a parse tree produced by CMakeParser#single_argument.
    def exitSingle_argument(self, ctx:CMakeParser.Single_argumentContext):
        pass


    # Enter a parse tree produced by CMakeParser#compound_argument.
    def enterCompound_argument(self, ctx:CMakeParser.Compound_argumentContext):
        pass

    # Exit a parse tree produced by CMakeParser#compound_argument.
    def exitCompound_argument(self, ctx:CMakeParser.Compound_argumentContext):
        pass





import logging

from antlr4 import Parser, DFA
from antlr4.atn.ATNConfigSet import ATNConfigSet
from antlr4.error.ErrorListener import ErrorListener


class CMakeSyntaxError(SyntaxError):
    pass


class ParserErrorListener(ErrorListener):
    """
    Listens for parser errors and raises exceptions when they occur.
    """

    def __init__(self):
        super(ParserErrorListener, self).__init__()

        self.logger = logging.getLogger(__name__)

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        """
        As per the Antlr4 Javadoc:

            Upon syntax error, notify any interested parties. This is not how to recover from errors or compute error
            messages. ErrorStrategy specifies how to recover from syntax errors and how to compute error messages.
            This listener's job is simply to emit a computed message, though it has enough information to create its
            own message in many cases. The RecognitionException (e) is non-null for all syntax errors except when we
            discover mismatched token errors that we can recover from in-line, without returning from the surrounding
            rule (via the single token insertion and deletion mechanism).

        :raises RecognitionException: When the mismatched token cannot be recovered from
        :raises CMakeSyntaxError: When it is possible to recover from this error via single token insertion/deletion.
        """
        if e is not None:
            raise e
        else:
            s = CMakeSyntaxError()
            s.lineno = f"{line}:{column}"
            s.msg = msg
            raise s

    def reportAmbiguity(self, recognizer: Parser, dfa: DFA, startIndex: int,
                        stopIndex: int, exact: bool, ambigAlts: set, configs: ATNConfigSet):
        """
        As per the Antlr4 Javadoc:

            This method is called by the parser when a full-context prediction results in an ambiguity. Each
            full-context prediction which does not result in a syntax error will call either
            ErrorListener.reportContextSensitivity() or ErrorListener.reportAmbiguity().

            When ambigAlts is not None, it contains the set of potentially viable alternatives identified by the
            prediction algorithm. When ambigAlts is None, use ATNConfigSet.getAlts() to obtain the represented
            alternatives from the configs argument.

            When exact is True, all of the potentially viable alternatives are truly viable, i.e. this is reporting
            an exact ambiguity. When exact is False, at least two of the potentially viable alternatives are viable
            for the current input, but the prediction algorithm terminated as soon as it determined that at least the
            minimum potentially viable alternative is truly viable.

            When the PredictionMode.LL_EXACT_AMBIG_DETECTION prediction mode is used, the parser is required to
            identify exact ambiguities so exact will always be true.

            This method is not used by lexers.

        This method only logs ambiguities, it does not halt parsing as our grammar
        is ambiguous. The chosen alternative is implied to be the first one in the grammar.
        """
        self.logger.debug(f"Found parse ambiguity at starting index {startIndex}, stop index {stopIndex}")

    def reportAttemptingFullContext(self, recognizer: Parser, dfa: DFA, startIndex: int,
                                    stopIndex: int, conflictingAlts: set, configs: ATNConfigSet):
        """
        As per the Antlr4 Javadoc:

            This method is called when an SLL conflict occurs and the parser is about to use the full context
            information to make an LL decision. If one or more configurations in configs contains a semantic
            predicate, the predicates are evaluated before this method is called. The subset of alternatives which
            are still viable after predicates are evaluated is reported in conflictingAlts.

            This method is not used by lexers.

        Currently, this is a no-op, it is unknown whether we should allow or disallow full context LL decisions.
        """
        pass

    def reportContextSensitivity(self, recognizer: Parser, dfa: DFA, startIndex: int,
                                 stopIndex: int, prediction: int, configs: ATNConfigSet):
        """
        As per the Antlr4 Javadoc:

            This method is called by the parser when a full-context prediction has a unique result. Each full-context
            prediction which does not result in a syntax error will call either
            ErrorListener.reportContextSensitivity() or ErrorListener.reportAmbiguity().

            For prediction implementations that only evaluate full-context predictions when an SLL conflict is found
            (including the default ParserATNSimulator implementation), this method reports cases where SLL conflicts
            were resolved to unique full-context predictions, i.e. the decision was context-sensitive. This report
            does not necessarily indicate a problem, and it may appear even in completely unambiguous grammars.
            configs may have more than one represented alternative if the full-context prediction algorithm does not
            evaluate predicates before beginning the full-context prediction. In all cases, the final prediction is
            passed as the prediction argument.

            Note that the definition of "context sensitivity" in this method differs from the concept in
            DecisionInfo.contextSensitivities. This method reports all instances where an SLL conflict occurred but
            LL parsing produced a unique result, whether or not that unique result matches the minimum alternative in
            the SLL conflicting set.

            This method is not used by lexers.

        This method is currently a no-op, as it being called does not indicate an error necessarily.
        """
        pass



# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This file contains the Documenter class, which is the top-level
entrypoint into the CMinx documentation system. It handles parsing,
aggregating, and RST writing.

:Author: Branden Butler
:License: Apache 2.0

"""
from typing import List

from antlr4 import *

from .aggregator import DocumentationAggregator
from cminx import Settings
from .documentation_types import DocumentationType, ModuleDocumentation
from .parser import ParserErrorListener
from .parser.CMakeLexer import CMakeLexer
from .parser.CMakeParser import CMakeParser
from .rstwriter import RSTWriter, Directive


class Documenter(object):
    """
    Generates RST documentation from aggregated documentation, combining
    :class:`~cminx.aggregator.DocumentationAggregator` and :class:`~cminx.rstwriter.RSTWriter`.
    The entrypoint for this class is the :meth:`process` method.
    """

    def __init__(
            self,
            file: str,
            title: str = None,
            module_name: str = None,
            settings: Settings = Settings()) -> None:
        """
        :param file: CMake file to read documentation from.
        :param title: RST header title to use in the generated document.
        :param module_name: The name of the CMake module, used in the module directive if
                            no other name is given in a module doccomment.
        :param settings: Dictionary containing application settings for documentation
        """

        self.settings: Settings = settings
        """Settings used for aggregation and RST generation, passed down to all downstream components."""

        title = file if title is None else title

        if module_name is None:
            module_name = title

        self.module_name: str = module_name
        """The name of the CMake module, used as a default when no name given via a module doccomment."""

        self.writer: RSTWriter = RSTWriter(title, settings=settings)
        """The writer that is passed to the aggregated documentation objects."""

        self.module: Directive
        """The :code:`.. module::` directive that defines the module's name."""

        # We need a string stream of some kind, FileStream is easiest
        self.input_stream: InputStream = FileStream(file)
        """The string stream used to read the CMake file."""

        # Convert those strings into tokens and build a stream from those
        self.lexer: CMakeLexer = CMakeLexer(self.input_stream)
        """The lexer used to generate the token stream."""

        self.stream: TokenStream = CommonTokenStream(self.lexer)
        """The stream of tokens from the lexer, should be passed to the parser."""

        # We now have a stream of CommonToken instead of strings, parsers
        # require this type of stream
        self.parser: CMakeParser = CMakeParser(self.stream)
        """
        The parser used to parse the token stream and call our listener.
        By default, an instance of :class:`~cminx.parser.ParserErrorListener` is
        added.
        """

        self.parser.addErrorListener(ParserErrorListener())

        # Hard part is done, we now have a fully usable parse tree, now we just
        # need to walk it
        self.aggregator: DocumentationAggregator = DocumentationAggregator(settings)
        """The aggregator used to listen for parser rules and generate documentation objects."""

        self.walker: ParseTreeWalker = ParseTreeWalker()
        """Walks the parser tree at a given parser rule, used to kick off aggregation."""

    def process(self) -> RSTWriter:
        """
        Process :attr:`cminx.aggregator.DocumentationAggregator.documented` and build RST document from it.

        :return: Completed RSTWriter document, also located in :attr:`Documenter.writer`
        """

        # Parse and lex the file, then walk the tree and aggregate the
        # documented commands
        self.walker.walk(self.aggregator, self.parser.cmake_file())

        # All the documented commands are now stored in aggregator.documented,
        # each element is a namedtuple representing the type of documentation it is.
        # So far we can document functions, macros, and variables (only strings
        # and lists built using set)
        self.process_docs(self.aggregator.documented)
        return self.writer

    def process_docs(self, docs: List[DocumentationType]) -> None:
        """
        Loops over document and calls :meth:`cminx.documentation_types.DocumentationType.process` with
        :attr:`~Documenter.writer`.

        :param docs: List of documentation objects.
        """

        module_docs = [x for x in docs if isinstance(x, ModuleDocumentation)]

        if len(module_docs) == 0:
            docs.insert(0, ModuleDocumentation(self.module_name, ""))

        for module_doc in module_docs:
            if module_doc.name is None or len(module_doc.name) == 0:
                module_doc.name = self.module_name
            else:
                self.writer.title = module_doc.name

        for doc in docs:
            doc.process(self.writer)



#!/usr/bin/python3
# Copyright 2021 CMakePP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from cminx.rstwriter import RSTWriter
#from cminx.rst_validator import RSTValidator #No longer in cminx package, included in test/ directory instead as not needed for general operation

if __name__ == "__main__":

    #First we need to build an RSTWriter to validate
    writer = RSTWriter("Title")
    writer.text("This is a test")
    writer.bulleted_list("test", "No")
    writer.enumerated_list(1, "two")
    writer.doctest('print("test")', 'test')
    d = writer.directive("DANGER")
    d.text("This is john")


    #Make a table of values
    tab = [
    [0, 1, 2, 3],
    [4, 5, 6, 7],
    [8, 9, 10, 11]
    ]
    writer.simple_table(tab, column_headings=["Column 1", "Column 2", "Column 3", "Column 4"])


    #Next step is very simple. Just instantiate the validator with your desired settings and the writer
    #Then call validate()
    validator = RSTValidator(writer, werror = True)
    print(validator.validate())
